# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import webtest
from testing_support.fixtures import dt_enabled
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_span_events import validate_span_events


@pytest.fixture(scope="session")
def graphql_wsgi_run():
    """Wrapper function to simulate framework_graphql test behavior."""
    from _target_application import _target_wsgi_application

    app = webtest.TestApp(_target_wsgi_application)

    def execute(query):
        return app.post_json("/", {"query": query})

    return execute


@dt_enabled
def test_query_and_mutation_wsgi(graphql_wsgi_run):
    from graphql import __version__ as version

    FRAMEWORK_METRICS = [
        ("Python/Framework/Ariadne/None", 1),
        ("Python/Framework/GraphQL/%s" % version, 1),
    ]
    _test_mutation_scoped_metrics = [
        ("GraphQL/resolve/Ariadne/storage_add", 1),
        ("GraphQL/operation/Ariadne/mutation/<anonymous>/storage_add.string", 1),
    ]
    _test_query_scoped_metrics = [
        ("GraphQL/resolve/Ariadne/storage", 1),
        ("GraphQL/operation/Ariadne/query/<anonymous>/storage", 1),
    ]
    _test_unscoped_metrics = [
        ("WebTransaction", 1),
        ("Python/WSGI/Response", 1),
        ("GraphQL/all", 1),
        ("GraphQL/Ariadne/all", 1),
        ("GraphQL/allWeb", 1),
        ("GraphQL/Ariadne/allWeb", 1),
    ]
    _test_mutation_unscoped_metrics = _test_unscoped_metrics + _test_mutation_scoped_metrics
    _test_query_unscoped_metrics = _test_unscoped_metrics + _test_query_scoped_metrics

    _expected_mutation_operation_attributes = {
        "graphql.operation.type": "mutation",
        "graphql.operation.name": "<anonymous>",
    }
    _expected_mutation_resolver_attributes = {
        "graphql.field.name": "storage_add",
        "graphql.field.parentType": "Mutation",
        "graphql.field.path": "storage_add",
        "graphql.field.returnType": "StorageAdd",
    }
    _expected_query_operation_attributes = {
        "graphql.operation.type": "query",
        "graphql.operation.name": "<anonymous>",
    }
    _expected_query_resolver_attributes = {
        "graphql.field.name": "storage",
        "graphql.field.parentType": "Query",
        "graphql.field.path": "storage",
        "graphql.field.returnType": "[String]",
    }

    @validate_transaction_metrics(
        "query/<anonymous>/storage",
        "GraphQL",
        scoped_metrics=_test_query_scoped_metrics,
        rollup_metrics=_test_query_unscoped_metrics + FRAMEWORK_METRICS,
    )
    @validate_transaction_metrics(
        "mutation/<anonymous>/storage_add.string",
        "GraphQL",
        scoped_metrics=_test_mutation_scoped_metrics,
        rollup_metrics=_test_mutation_unscoped_metrics + FRAMEWORK_METRICS,
        index=-2,
    )
    @validate_span_events(exact_agents=_expected_mutation_operation_attributes, index=-2)
    @validate_span_events(exact_agents=_expected_mutation_resolver_attributes, index=-2)
    @validate_span_events(exact_agents=_expected_query_operation_attributes)
    @validate_span_events(exact_agents=_expected_query_resolver_attributes)
    def _test():
        response = graphql_wsgi_run('mutation { storage_add(string: "abc") { string } }')
        assert response.status_code == 200
        response = response.json_body
        assert not response.get("errors")

        response = graphql_wsgi_run("query { storage }")
        assert response.status_code == 200
        response = response.json_body
        assert not response.get("errors")

        # These are separate assertions because pypy stores 'abc' as a unicode string while other Python versions do not
        assert "storage" in str(response.get("data"))
        assert "abc" in str(response.get("data"))

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections

import pytest

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

from newrelic.common.object_names import callable_name
from newrelic.config import delete_setting, translate_deprecated_settings
from newrelic.core.config import (
    Settings,
    apply_config_setting,
    apply_server_side_settings,
    fetch_config_setting,
    flatten_settings,
    global_settings,
    global_settings_dump,
)


def parameterize_local_config(settings_list):
    settings_object_list = []

    for settings in settings_list:
        settings_object = Settings()
        for name, value in settings.items():
            apply_config_setting(settings_object, name, value)
        settings_object_list.append(settings_object)

    return pytest.mark.parametrize("settings", settings_object_list)


_test_dictionary_local_config = [
    {
        "request_headers_map": {"NR-SESSION": "BLANK"},
        "event_harvest_config": {
            "harvest_limits": {
                "analytic_event_data": 100,
                "custom_event_data": 100,
                "error_event_data": 8,
                "log_event_data": 100,
            },
            "report_period_ms": 5000,
        },
    }
]

_test_strip_proxy_details_local_configs = [
    {
        "license_key": "LICENSE-KEY",
        "api_key": "API-KEY",
        "encoding_key": "ENCODING-KEY",
        "js_agent_loader": "AGENT-LOADER-JS",
        "js_agent_file": "AGENT-LOADER-FILE",
        "proxy_scheme": None,
        "proxy_host": None,
        "proxy_port": None,
        "proxy_user": None,
        "proxy_pass": None,
        "expected_proxy_host": None,
    },
    {
        "license_key": "LICENSE-KEY",
        "api_key": "API-KEY",
        "encoding_key": "ENCODING-KEY",
        "js_agent_loader": "AGENT-LOADER-JS",
        "js_agent_file": "AGENT-LOADER-FILE",
        "proxy_scheme": None,
        "proxy_host": "hostname",
        "proxy_port": 8888,
        "proxy_user": None,
        "proxy_pass": None,
        "expected_proxy_host": "hostname",
    },
    {
        "license_key": "LICENSE-KEY",
        "api_key": "API-KEY",
        "encoding_key": "ENCODING-KEY",
        "js_agent_loader": "AGENT-LOADER-JS",
        "js_agent_file": "AGENT-LOADER-FILE",
        "proxy_scheme": None,
        "proxy_host": "hostname",
        "proxy_port": 8888,
        "proxy_user": "username",
        "proxy_pass": "password",
        "expected_proxy_host": "hostname",
    },
    {
        "license_key": "LICENSE-KEY",
        "api_key": "API-KEY",
        "encoding_key": "ENCODING-KEY",
        "js_agent_loader": "AGENT-LOADER-JS",
        "js_agent_file": "AGENT-LOADER-FILE",
        "proxy_scheme": None,
        "proxy_host": "hostname",
        "proxy_port": 8888,
        "proxy_user": "username",
        "proxy_pass": None,
        "expected_proxy_host": "hostname",
    },
    {
        "license_key": "LICENSE-KEY",
        "api_key": "API-KEY",
        "encoding_key": "ENCODING-KEY",
        "js_agent_loader": "AGENT-LOADER-JS",
        "js_agent_file": "AGENT-LOADER-FILE",
        "proxy_scheme": None,
        "proxy_host": "hostname",
        "proxy_port": 8888,
        "proxy_user": "username",
        "proxy_pass": "",
        "expected_proxy_host": "hostname",
    },
    {
        "license_key": "LICENSE-KEY",
        "api_key": "API-KEY",
        "encoding_key": "ENCODING-KEY",
        "js_agent_loader": "AGENT-LOADER-JS",
        "js_agent_file": "AGENT-LOADER-FILE",
        "proxy_scheme": None,
        "proxy_host": "http://hostname",
        "proxy_port": None,
        "proxy_user": None,
        "proxy_pass": None,
        "expected_proxy_host": "http://hostname",
    },
    {
        "license_key": "LICENSE-KEY",
        "api_key": "API-KEY",
        "encoding_key": "ENCODING-KEY",
        "js_agent_loader": "AGENT-LOADER-JS",
        "js_agent_file": "AGENT-LOADER-FILE",
        "proxy_scheme": None,
        "proxy_host": "http://hostname:8888",
        "proxy_port": None,
        "proxy_user": None,
        "proxy_pass": None,
        "expected_proxy_host": "http://hostname:8888",
    },
    {
        "license_key": "LICENSE-KEY",
        "api_key": "API-KEY",
        "encoding_key": "ENCODING-KEY",
        "js_agent_loader": "AGENT-LOADER-JS",
        "js_agent_file": "AGENT-LOADER-FILE",
        "proxy_scheme": None,
        "proxy_host": "http://hostname",
        "proxy_port": None,
        "proxy_user": "username",
        "proxy_pass": "password",
        "expected_proxy_host": "http://hostname",
    },
    {
        "license_key": "LICENSE-KEY",
        "api_key": "API-KEY",
        "encoding_key": "ENCODING-KEY",
        "js_agent_loader": "AGENT-LOADER-JS",
        "js_agent_file": "AGENT-LOADER-FILE",
        "proxy_scheme": None,
        "proxy_host": "http://hostname:8888",
        "proxy_port": None,
        "proxy_user": "username",
        "proxy_pass": "password",
        "expected_proxy_host": "http://hostname:8888",
    },
    {
        "license_key": "LICENSE-KEY",
        "api_key": "API-KEY",
        "encoding_key": "ENCODING-KEY",
        "js_agent_loader": "AGENT-LOADER-JS",
        "js_agent_file": "AGENT-LOADER-FILE",
        "proxy_scheme": None,
        "proxy_host": "http://username@hostname",
        "proxy_port": None,
        "proxy_user": None,
        "proxy_pass": None,
        "expected_proxy_host": "http://****@hostname",
    },
    {
        "license_key": "LICENSE-KEY",
        "api_key": "API-KEY",
        "encoding_key": "ENCODING-KEY",
        "js_agent_loader": "AGENT-LOADER-JS",
        "js_agent_file": "AGENT-LOADER-FILE",
        "proxy_scheme": None,
        "proxy_host": "http://username@hostname:8888",
        "proxy_port": None,
        "proxy_user": None,
        "proxy_pass": None,
        "expected_proxy_host": "http://****@hostname:8888",
    },
    {
        "license_key": "LICENSE-KEY",
        "api_key": "API-KEY",
        "encoding_key": "ENCODING-KEY",
        "js_agent_loader": "AGENT-LOADER-JS",
        "js_agent_file": "AGENT-LOADER-FILE",
        "proxy_scheme": None,
        "proxy_host": "http://username:password@hostname",
        "proxy_port": None,
        "proxy_user": None,
        "proxy_pass": None,
        "expected_proxy_host": "http://****:****@hostname",
    },
    {
        "license_key": "LICENSE-KEY",
        "api_key": "API-KEY",
        "encoding_key": "ENCODING-KEY",
        "js_agent_loader": "AGENT-LOADER-JS",
        "js_agent_file": "AGENT-LOADER-FILE",
        "proxy_scheme": None,
        "proxy_host": "http://username:password@hostname:8888",
        "proxy_port": None,
        "proxy_user": None,
        "proxy_pass": None,
        "expected_proxy_host": "http://****:****@hostname:8888",
    },
    {
        "license_key": "LICENSE-KEY",
        "api_key": "API-KEY",
        "encoding_key": "ENCODING-KEY",
        "js_agent_loader": "AGENT-LOADER-JS",
        "js_agent_file": "AGENT-LOADER-FILE",
        "proxy_scheme": None,
        "proxy_host": "https://username:password@hostname",
        "proxy_port": None,
        "proxy_user": None,
        "proxy_pass": None,
        "expected_proxy_host": "https://****:****@hostname",
    },
    {
        "license_key": "LICENSE-KEY",
        "api_key": "API-KEY",
        "encoding_key": "ENCODING-KEY",
        "js_agent_loader": "AGENT-LOADER-JS",
        "js_agent_file": "AGENT-LOADER-FILE",
        "proxy_scheme": None,
        "proxy_host": "https://username:password@hostname:8888",
        "proxy_port": None,
        "proxy_user": None,
        "proxy_pass": None,
        "expected_proxy_host": "https://****:****@hostname:8888",
    },
]


@parameterize_local_config(_test_dictionary_local_config)
def test_dict_parse(settings):

    assert "NR-SESSION" in settings.request_headers_map

    config = settings.event_harvest_config

    assert "harvest_limits" in config
    assert "report_period_ms" in config

    limits = config.harvest_limits

    assert "analytic_event_data" in limits
    assert "custom_event_data" in limits
    assert "error_event_data" in limits
    assert "log_event_data" in limits


@parameterize_local_config(_test_strip_proxy_details_local_configs)
def test_strip_proxy_details(settings):
    assert "license_key" in settings
    assert "api_key" in settings

    assert "proxy_scheme" in settings
    assert "proxy_host" in settings
    assert "proxy_port" in settings
    assert "proxy_user" in settings
    assert "proxy_port" in settings

    stripped = global_settings_dump(settings)

    # These should always be deleted.

    assert "license_key" not in stripped
    assert "api_key" not in stripped

    # These should be obfuscated or None.

    obfuscated = "****"

    assert stripped["proxy_user"] in (None, obfuscated)
    assert stripped["proxy_pass"] in (None, obfuscated)

    # The proxy_host and proxy_port will be preserved but proxy_host
    # needs to be checked to make sure it doesn't contain a username and
    # password. Do this through parsing to make sure not breaking the
    # structure in some way and also direct comparison of what is
    # expected as backup.

    assert "proxy_scheme" in settings
    assert "proxy_host" in stripped
    assert "proxy_port" in stripped

    proxy_host = stripped["proxy_host"]
    expected_proxy_host = stripped["expected_proxy_host"]

    if proxy_host is not None:
        components = urlparse.urlparse(proxy_host)

        if components.username:
            assert components.username == obfuscated
        if components.password:
            assert components.password == obfuscated

    assert proxy_host == expected_proxy_host


def test_delete_setting():
    d = {"transaction_tracer.capture_attributes": True}
    settings = apply_server_side_settings(d)
    assert "capture_attributes" in settings.transaction_tracer

    delete_setting(settings, "transaction_tracer.capture_attributes")
    assert "capture_attributes" not in settings.transaction_tracer


def test_delete_setting_absent():
    settings = apply_server_side_settings()
    assert "capture_attributes" not in settings.transaction_tracer

    delete_setting(settings, "transaction_tracer.capture_attributes")
    assert "capture_attributes" not in settings.transaction_tracer


def test_delete_setting_parent():
    settings = apply_server_side_settings()
    assert "transaction_tracer" in settings
    assert "attributes" in settings.transaction_tracer

    delete_setting(settings, "transaction_tracer")
    assert "transaction_tracer" not in settings


# Build a series of tests for `translate_deprecated_setting`.
#
# Each test will consist of an old setting and new setting.
# For each setting, there are 2 variations:
#
#       'value' == 'default'
#       'value' != 'default'
TSetting = collections.namedtuple("TSetting", ["name", "value", "default"])

translate_settings_tests = [
    (
        TSetting("strip_exception_messages.whitelist", [], []),
        TSetting("strip_exception_messages.allowlist", ["non-default-value"], []),
    ),
    (
        TSetting("strip_exception_messages.whitelist", ["non-default-value"], []),
        TSetting("strip_exception_messages.allowlist", [], []),
    ),
    (
        TSetting("transaction_tracer.capture_attributes", True, True),
        TSetting("transaction_tracer.attributes.enabled", False, True),
    ),
    (
        TSetting("transaction_tracer.capture_attributes", False, True),
        TSetting("transaction_tracer.attributes.enabled", True, True),
    ),
    (
        TSetting("error_collector.capture_attributes", True, True),
        TSetting("error_collector.attributes.enabled", False, True),
    ),
    (
        TSetting("error_collector.capture_attributes", False, True),
        TSetting("error_collector.attributes.enabled", True, True),
    ),
    (
        TSetting("browser_monitoring.capture_attributes", False, False),
        TSetting("browser_monitoring.attributes.enabled", True, False),
    ),
    (
        TSetting("browser_monitoring.capture_attributes", True, False),
        TSetting("browser_monitoring.attributes.enabled", False, False),
    ),
    (
        TSetting("analytics_events.capture_attributes", True, True),
        TSetting("transaction_events.attributes.enabled", False, True),
    ),
    (
        TSetting("analytics_events.capture_attributes", False, True),
        TSetting("transaction_events.attributes.enabled", True, True),
    ),
    (TSetting("analytics_events.enabled", True, True), TSetting("transaction_events.enabled", False, True)),
    (TSetting("analytics_events.enabled", False, True), TSetting("transaction_events.enabled", True, True)),
    (
        TSetting("analytics_events.max_samples_stored", 1200, 1200),
        TSetting("event_harvest_config.harvest_limits.analytic_event_data", 9999, 1200),
    ),
    (
        TSetting("analytics_events.max_samples_stored", 9999, 1200),
        TSetting("event_harvest_config.harvest_limits.analytic_event_data", 1200, 1200),
    ),
    (
        TSetting("transaction_events.max_samples_stored", 1200, 1200),
        TSetting("event_harvest_config.harvest_limits.analytic_event_data", 9999, 1200),
    ),
    (
        TSetting("transaction_events.max_samples_stored", 9999, 1200),
        TSetting("event_harvest_config.harvest_limits.analytic_event_data", 1200, 1200),
    ),
    (
        TSetting("span_events.max_samples_stored", 1000, 2000),
        TSetting("event_harvest_config.harvest_limits.span_event_data", 9999, 2000),
    ),
    (
        TSetting("span_events.max_samples_stored", 9999, 2000),
        TSetting("event_harvest_config.harvest_limits.span_event_data", 1000, 2000),
    ),
    (
        TSetting("error_collector.max_event_samples_stored", 100, 100),
        TSetting("event_harvest_config.harvest_limits.error_event_data", 9999, 100),
    ),
    (
        TSetting("error_collector.max_event_samples_stored", 9999, 100),
        TSetting("event_harvest_config.harvest_limits.error_event_data", 100, 100),
    ),
    (
        TSetting("custom_insights_events.max_samples_stored", 3600, 3600),
        TSetting("event_harvest_config.harvest_limits.custom_event_data", 9999, 3600),
    ),
    (
        TSetting("custom_insights_events.max_samples_stored", 9999, 3600),
        TSetting("event_harvest_config.harvest_limits.custom_event_data", 3600, 3600),
    ),
    (
        TSetting("application_logging.forwarding.max_samples_stored", 10000, 10000),
        TSetting("event_harvest_config.harvest_limits.log_event_data", 99999, 10000),
    ),
    (
        TSetting("application_logging.forwarding.max_samples_stored", 99999, 10000),
        TSetting("event_harvest_config.harvest_limits.log_event_data", 10000, 10000),
    ),
    (
        TSetting("error_collector.ignore_errors", [], []),
        TSetting("error_collector.ignore_classes", callable_name(ValueError), []),
    ),
    (
        TSetting("error_collector.ignore_errors", callable_name(ValueError), []),
        TSetting("error_collector.ignore_classes", [], []),
    ),
]


@pytest.mark.parametrize("old,new", translate_settings_tests)
def test_translate_deprecated_setting_without_new_setting(old, new):
    # Before: deprecated setting will be in settings object.
    #         new setting will be in settings object and have default value
    #
    # After:  deprecated setting will *NOT* be in settings object
    #         new setting will have value of deprecated setting

    settings = apply_server_side_settings()
    apply_config_setting(settings, old.name, old.value)

    assert fetch_config_setting(settings, old.name) == old.value
    assert fetch_config_setting(settings, new.name) == new.default

    cached = [(old.name, old.value)]
    result = translate_deprecated_settings(settings, cached)

    assert result is settings
    assert old.name not in flatten_settings(result)
    assert fetch_config_setting(result, new.name) == old.value


@pytest.mark.parametrize("old,new", translate_settings_tests)
def test_translate_deprecated_setting_with_new_setting(old, new):
    # Before: deprecated setting will be in settings object.
    #         new setting will be in settings object and have its value
    #
    # After:  deprecated setting will *NOT* be in settings object
    #         new setting will still have its value (remain unchanged)

    settings = apply_server_side_settings()
    apply_config_setting(settings, old.name, old.value)
    apply_config_setting(settings, new.name, new.value)

    assert fetch_config_setting(settings, old.name) == old.value
    assert fetch_config_setting(settings, new.name) == new.value

    cached = [(old.name, old.value), (new.name, new.value)]
    result = translate_deprecated_settings(settings, cached)

    assert result is settings
    assert old.name not in flatten_settings(result)
    assert fetch_config_setting(result, new.name) == new.value


@pytest.mark.parametrize("old,new", translate_settings_tests)
def test_translate_deprecated_setting_without_old_setting(old, new):
    # Before: deprecated setting will *NOT* be in settings object.
    #         new setting will be in settings object and have its value
    #
    # After:  deprecated setting will still *NOT* be in settings object
    #         new setting will still have its value (remain unchanged)

    settings = apply_server_side_settings()
    apply_config_setting(settings, new.name, new.value)

    assert old.name not in flatten_settings(settings)
    assert fetch_config_setting(settings, new.name) == new.value

    cached = [(new.name, new.value)]
    result = translate_deprecated_settings(settings, cached)

    assert result is settings
    assert old.name not in flatten_settings(result)
    assert fetch_config_setting(result, new.name) == new.value


def test_translate_deprecated_ignored_params_without_new_setting():
    ignored_params = ["foo", "bar"]
    settings = apply_server_side_settings()
    apply_config_setting(settings, "ignored_params", ignored_params)

    assert "foo" in settings.ignored_params
    assert "bar" in settings.ignored_params
    assert len(settings.attributes.exclude) == 0

    cached = [("ignored_params", ignored_params)]
    result = translate_deprecated_settings(settings, cached)

    assert result is settings
    assert "request.parameters.foo" in result.attributes.exclude
    assert "request.parameters.bar" in result.attributes.exclude
    assert "ignored_params" not in result


def test_translate_deprecated_ignored_params_with_new_setting():
    ignored_params = ["foo", "bar"]
    attr_exclude = ["request.parameters.foo"]
    settings = apply_server_side_settings()
    apply_config_setting(settings, "ignored_params", ignored_params)
    apply_config_setting(settings, "attributes.exclude", attr_exclude)

    assert "foo" in settings.ignored_params
    assert "bar" in settings.ignored_params
    assert "request.parameters.foo" in settings.attributes.exclude

    cached = [("ignored_params", ignored_params), ("attributes.exclude", attr_exclude)]
    result = translate_deprecated_settings(settings, cached)

    # ignored_params are not merged!

    assert result is settings
    assert "request.parameters.foo" in result.attributes.exclude
    assert "request.parameters.bar" not in result.attributes.exclude
    assert "ignored_params" not in result


@pytest.mark.parametrize(
    "name,expected_value",
    (
        ("agent_run_id", None),
        ("entity_guid", None),
        ("distributed_tracing.exclude_newrelic_header", False),
    ),
)
def test_default_values(name, expected_value):
    settings = global_settings()
    value = fetch_config_setting(settings, name)
    assert value == expected_value



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import requests
import requests.exceptions
from testing_support.external_fixtures import (
    cache_outgoing_headers,
    insert_incoming_headers,
)
from testing_support.fixtures import (
    cat_enabled,
    override_application_settings,
    validate_tt_parenting,
)
from testing_support.validators.validate_cross_process_headers import (
    validate_cross_process_headers,
)
from testing_support.validators.validate_external_node_params import (
    validate_external_node_params,
)
from testing_support.validators.validate_transaction_errors import validate_transaction_errors
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from newrelic.api.background_task import background_task


def get_requests_version():
    return tuple(map(int, requests.__version__.split(".")[:2]))


@pytest.fixture(scope="session")
def metrics(server):
    scoped = [("External/localhost:%d/requests/" % server.port, 1)]

    rollup = [
        ("External/all", 1),
        ("External/allOther", 1),
        ("External/localhost:%d/all" % server.port, 1),
        ("External/localhost:%d/requests/" % server.port, 1),
    ]

    return scoped, rollup


_test_requests_http_request_get_parenting = (
    "TransactionNode",
    [
        ("ExternalNode", []),
    ],
)


def test_http_request_get(server, metrics):
    @validate_tt_parenting(_test_requests_http_request_get_parenting)
    @validate_transaction_errors(errors=[])
    @validate_transaction_metrics(
        "test_requests:test_http_request_get",
        scoped_metrics=metrics[0],
        rollup_metrics=metrics[1],
        background_task=True,
    )
    @background_task(name="test_requests:test_http_request_get")
    def _test():
        requests.get("http://localhost:%d/" % server.port)

    _test()


@pytest.mark.skipif(get_requests_version() < (0, 8), reason="Can't set verify=False for requests.get() in v0.7")
def test_https_request_get(server, metrics):
    @validate_transaction_errors(errors=[])
    @validate_transaction_metrics(
        "test_requests:test_https_request_get",
        scoped_metrics=metrics[0],
        rollup_metrics=metrics[1],
        background_task=True,
    )
    @background_task(name="test_requests:test_https_request_get")
    def _test():
        try:
            requests.get("https://localhost:%d/" % server.port, verify=False)
        except Exception:
            pass

    _test()


@pytest.mark.skipif(get_requests_version() < (1, 0), reason="Session.send() doesn't exist for requests < v1.0.")
def test_http_session_send(server, metrics):
    @validate_transaction_errors(errors=[])
    @validate_transaction_metrics(
        "test_requests:test_http_session_send",
        scoped_metrics=metrics[0],
        rollup_metrics=metrics[1],
        background_task=True,
    )
    @background_task(name="test_requests:test_http_session_send")
    def _test():
        session = requests.Session()
        req = requests.Request("GET", "http://localhost:%d/" % server.port)
        prep_req = req.prepare()
        session.send(prep_req)

    _test()


_test_requests_none_url_scoped_metrics = [("External/unknown/requests/", 1)]

_test_requests_none_url_rollup_metrics = [
    ("External/all", 1),
    ("External/allOther", 1),
    ("External/unknown/all", 1),
    ("External/unknown/requests/", 1),
]


@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    "test_requests:test_none_url_get",
    scoped_metrics=_test_requests_none_url_scoped_metrics,
    rollup_metrics=_test_requests_none_url_rollup_metrics,
    background_task=True,
)
@background_task()
def test_none_url_get():
    try:
        requests.get(None)
    except requests.exceptions.MissingSchema:
        # Python 2.
        pass
    except TypeError:
        # Python 3.
        pass


_test_requests_wrong_datatype_url_scoped_metrics = [("External/unknown.url/requests/", 1)]

_test_requests_wrong_datatype_url_rollup_metrics = [
    ("External/all", 1),
    ("External/allOther", 1),
    ("External/unknown.url/all", 1),
    ("External/unknown.url/requests/", 1),
]


@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    "test_requests:test_wrong_datatype_url_get",
    scoped_metrics=_test_requests_wrong_datatype_url_scoped_metrics,
    rollup_metrics=_test_requests_wrong_datatype_url_rollup_metrics,
    background_task=True,
)
@background_task()
def test_wrong_datatype_url_get():
    try:
        requests.get({"a": 1})
    except Exception:
        pass


@pytest.mark.parametrize(
    "distributed_tracing,span_events",
    (
        (True, True),
        (True, False),
        (False, False),
    ),
)
def test_requests_cross_process_request(distributed_tracing, span_events, server):
    @validate_transaction_errors(errors=[])
    @background_task(name="test_requests:test_requests_cross_process_request")
    @cache_outgoing_headers
    @validate_cross_process_headers
    def _test():
        requests.get("http://localhost:%d/" % server.port)

    _test = override_application_settings(
        {
            "distributed_tracing.enabled": distributed_tracing,
            "cross_application_tracer.enabled": not distributed_tracing,
            "span_events.enabled": span_events,
        }
    )(_test)

    _test()


@cat_enabled
def test_requests_cross_process_response(server):
    _test_requests_cross_process_response_scoped_metrics = [
        ("ExternalTransaction/localhost:%d/1#2/test" % server.port, 1)
    ]

    _test_requests_cross_process_response_rollup_metrics = [
        ("External/all", 1),
        ("External/allOther", 1),
        ("External/localhost:%d/all" % server.port, 1),
        ("ExternalApp/localhost:%d/1#2/all" % server.port, 1),
        ("ExternalTransaction/localhost:%d/1#2/test" % server.port, 1),
    ]

    _test_requests_cross_process_response_external_node_params = [
        ("cross_process_id", "1#2"),
        ("external_txn_name", "test"),
        ("transaction_guid", "0123456789012345"),
    ]

    @validate_transaction_errors(errors=[])
    @validate_transaction_metrics(
        "test_requests:test_requests_cross_process_response",
        scoped_metrics=_test_requests_cross_process_response_scoped_metrics,
        rollup_metrics=_test_requests_cross_process_response_rollup_metrics,
        background_task=True,
    )
    @insert_incoming_headers
    @validate_external_node_params(params=_test_requests_cross_process_response_external_node_params)
    @background_task(name="test_requests:test_requests_cross_process_response")
    def _test():
        requests.get("http://localhost:%d/" % server.port)

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import psycopg2cffi
import psycopg2cffi.extensions
import psycopg2cffi.extras

from testing_support.fixtures import validate_stats_engine_explain_plan_output_is_none
from testing_support.validators.validate_transaction_errors import validate_transaction_errors
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_transaction_slow_sql_count import \
        validate_transaction_slow_sql_count
from testing_support.validators.validate_database_trace_inputs import validate_database_trace_inputs

from testing_support.db_settings import postgresql_settings

from newrelic.api.background_task import background_task

DB_SETTINGS = postgresql_settings()[0]

_test_execute_via_cursor_scoped_metrics = [
        ('Function/psycopg2cffi:connect', 1),
        ('Function/psycopg2cffi._impl.connection:Connection.__enter__', 1),
        ('Function/psycopg2cffi._impl.connection:Connection.__exit__', 1),
        ('Datastore/statement/Postgres/%s/select' % DB_SETTINGS["table_name"], 1),
        ('Datastore/statement/Postgres/%s/insert' % DB_SETTINGS["table_name"], 1),
        ('Datastore/statement/Postgres/%s/update' % DB_SETTINGS["table_name"], 1),
        ('Datastore/statement/Postgres/%s/delete' % DB_SETTINGS["table_name"], 1),
        ('Datastore/statement/Postgres/now/call', 1),
        ('Datastore/statement/Postgres/pg_sleep/call', 1),
        ('Datastore/operation/Postgres/drop', 1),
        ('Datastore/operation/Postgres/create', 1),
        ('Datastore/operation/Postgres/commit', 3),
        ('Datastore/operation/Postgres/rollback', 1)]

_test_execute_via_cursor_rollup_metrics = [
        ('Datastore/all', 13),
        ('Datastore/allOther', 13),
        ('Datastore/Postgres/all', 13),
        ('Datastore/Postgres/allOther', 13),
        ('Datastore/operation/Postgres/select', 1),
        ('Datastore/statement/Postgres/%s/select' % DB_SETTINGS["table_name"], 1),
        ('Datastore/operation/Postgres/insert', 1),
        ('Datastore/statement/Postgres/%s/insert' % DB_SETTINGS["table_name"], 1),
        ('Datastore/operation/Postgres/update', 1),
        ('Datastore/statement/Postgres/%s/update' % DB_SETTINGS["table_name"], 1),
        ('Datastore/operation/Postgres/delete', 1),
        ('Datastore/statement/Postgres/%s/delete' % DB_SETTINGS["table_name"], 1),
        ('Datastore/operation/Postgres/drop', 1),
        ('Datastore/operation/Postgres/create', 1),
        ('Datastore/statement/Postgres/now/call', 1),
        ('Datastore/statement/Postgres/pg_sleep/call', 1),
        ('Datastore/operation/Postgres/call', 2),
        ('Datastore/operation/Postgres/commit', 3),
        ('Datastore/operation/Postgres/rollback', 1)]


@validate_transaction_metrics('test_database:test_execute_via_cursor',
        scoped_metrics=_test_execute_via_cursor_scoped_metrics,
        rollup_metrics=_test_execute_via_cursor_rollup_metrics,
        background_task=True)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_execute_via_cursor():
    with psycopg2cffi.connect(
            database=DB_SETTINGS['name'], user=DB_SETTINGS['user'],
            password=DB_SETTINGS['password'], host=DB_SETTINGS['host'],
            port=DB_SETTINGS['port']) as connection:

        cursor = connection.cursor()

        psycopg2cffi.extensions.register_type(psycopg2cffi.extensions.UNICODE)
        psycopg2cffi.extensions.register_type(
            psycopg2cffi.extensions.UNICODE,
            connection)
        psycopg2cffi.extensions.register_type(
            psycopg2cffi.extensions.UNICODE,
            cursor)

        cursor.execute("""drop table if exists %s""" % DB_SETTINGS["table_name"])

        cursor.execute("""create table %s """ % DB_SETTINGS["table_name"] +
                """(a integer, b real, c text)""")

        cursor.executemany("""insert into %s """ % DB_SETTINGS["table_name"] +
                """values (%s, %s, %s)""", [(1, 1.0, '1.0'),
                (2, 2.2, '2.2'), (3, 3.3, '3.3')])

        cursor.execute("""select * from %s""" % DB_SETTINGS["table_name"])

        for row in cursor:
            pass

        cursor.execute("""update %s""" % DB_SETTINGS["table_name"] + """ set a=%s, b=%s, """
                """c=%s where a=%s""", (4, 4.0, '4.0', 1))

        cursor.execute("""delete from %s where a=2""" % DB_SETTINGS["table_name"])

        connection.commit()

        cursor.callproc('now')
        cursor.callproc('pg_sleep', (0,))

        connection.rollback()
        connection.commit()


_test_rollback_on_exception_scoped_metrics = [
        ('Function/psycopg2cffi:connect', 1),
        ('Function/psycopg2cffi._impl.connection:Connection.__enter__', 1),
        ('Function/psycopg2cffi._impl.connection:Connection.__exit__', 1),
        ('Datastore/operation/Postgres/rollback', 1)]


_test_rollback_on_exception_rollup_metrics = [
        ('Datastore/all', 2),
        ('Datastore/allOther', 2),
        ('Datastore/Postgres/all', 2),
        ('Datastore/Postgres/allOther', 2)]


@validate_transaction_metrics('test_database:test_rollback_on_exception',
        scoped_metrics=_test_rollback_on_exception_scoped_metrics,
        rollup_metrics=_test_rollback_on_exception_rollup_metrics,
        background_task=True)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_rollback_on_exception():
    try:
        with psycopg2cffi.connect(
                database=DB_SETTINGS['name'], user=DB_SETTINGS['user'],
                password=DB_SETTINGS['password'], host=DB_SETTINGS['host'],
                port=DB_SETTINGS['port']):

            raise RuntimeError('error')
    except RuntimeError:
        pass


_test_async_mode_scoped_metrics = [
        ('Function/psycopg2cffi:connect', 1),
        ('Datastore/statement/Postgres/%s/select' % DB_SETTINGS["table_name"], 1),
        ('Datastore/statement/Postgres/%s/insert' % DB_SETTINGS["table_name"], 1),
        ('Datastore/operation/Postgres/drop', 1),
        ('Datastore/operation/Postgres/create', 1)]

_test_async_mode_rollup_metrics = [
        ('Datastore/all', 5),
        ('Datastore/allOther', 5),
        ('Datastore/Postgres/all', 5),
        ('Datastore/Postgres/allOther', 5),
        ('Datastore/operation/Postgres/select', 1),
        ('Datastore/statement/Postgres/%s/select' % DB_SETTINGS["table_name"], 1),
        ('Datastore/operation/Postgres/insert', 1),
        ('Datastore/statement/Postgres/%s/insert' % DB_SETTINGS["table_name"], 1),
        ('Datastore/operation/Postgres/drop', 1),
        ('Datastore/operation/Postgres/create', 1)]


@validate_stats_engine_explain_plan_output_is_none()
@validate_transaction_slow_sql_count(num_slow_sql=4)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@validate_transaction_metrics('test_database:test_async_mode',
        scoped_metrics=_test_async_mode_scoped_metrics,
        rollup_metrics=_test_async_mode_rollup_metrics,
        background_task=True)
@validate_transaction_errors(errors=[])
@background_task()
def test_async_mode():

    wait = psycopg2cffi.extras.wait_select

    kwargs = {}
    version = tuple(int(_) for _ in psycopg2cffi.__version__.split('.'))
    if version >= (2, 8):
        kwargs['async_'] = 1
    else:
        kwargs['async'] = 1

    async_conn = psycopg2cffi.connect(
            database=DB_SETTINGS['name'], user=DB_SETTINGS['user'],
            password=DB_SETTINGS['password'], host=DB_SETTINGS['host'],
            port=DB_SETTINGS['port'], **kwargs
    )
    wait(async_conn)
    async_cur = async_conn.cursor()

    async_cur.execute("""drop table if exists %s""" % DB_SETTINGS["table_name"])
    wait(async_cur.connection)

    async_cur.execute("""create table %s """ % DB_SETTINGS["table_name"] + 
            """(a integer, b real, c text)""")
    wait(async_cur.connection)

    async_cur.execute("""insert into %s """ % DB_SETTINGS["table_name"] + 
        """values (%s, %s, %s)""", (1, 1.0, '1.0'))
    wait(async_cur.connection)

    async_cur.execute("""select * from %s""" % DB_SETTINGS["table_name"])
    wait(async_cur.connection)

    for row in async_cur:
        assert isinstance(row, tuple)

    async_conn.close()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

import pika
import pytest
import six
import tornado
from compat import basic_consume
from conftest import (
    BODY,
    CORRELATION_ID,
    EXCHANGE,
    EXCHANGE_2,
    HEADERS,
    QUEUE,
    QUEUE_2,
    REPLY_TO,
)
from minversion import pika_version_info
from pika.adapters.tornado_connection import TornadoConnection
from testing_support.db_settings import rabbitmq_settings
from testing_support.fixtures import (
    capture_transaction_metrics,
    function_not_called,
    override_application_settings,
)
from testing_support.validators.validate_code_level_metrics import (
    validate_code_level_metrics,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)
from testing_support.validators.validate_tt_collector_json import (
    validate_tt_collector_json,
)

from newrelic.api.background_task import background_task

DB_SETTINGS = rabbitmq_settings()[0]

_message_broker_tt_params = {
    "queue_name": QUEUE,
    "routing_key": QUEUE,
    "correlation_id": CORRELATION_ID,
    "reply_to": REPLY_TO,
    "headers": HEADERS.copy(),
}


# Tornado's IO loop is not configurable in versions 5.x and up
try:

    class MyIOLoop(tornado.ioloop.IOLoop.configured_class()):
        def handle_callback_exception(self, *args, **kwargs):
            raise

    tornado.ioloop.IOLoop.configure(MyIOLoop)
except AttributeError:
    pass

connection_classes = [pika.SelectConnection, TornadoConnection]

parametrized_connection = pytest.mark.parametrize("ConnectionClass", connection_classes)


_test_select_conn_basic_get_inside_txn_metrics = [
    ("MessageBroker/RabbitMQ/Exchange/Produce/Named/%s" % EXCHANGE, None),
    ("MessageBroker/RabbitMQ/Exchange/Consume/Named/%s" % EXCHANGE, 1),
]

if six.PY3:
    _test_select_conn_basic_get_inside_txn_metrics.append(
        (
            (
                "Function/test_pika_async_connection_consume:"
                "test_async_connection_basic_get_inside_txn."
                "<locals>.on_message"
            ),
            1,
        )
    )
else:
    _test_select_conn_basic_get_inside_txn_metrics.append(("Function/test_pika_async_connection_consume:on_message", 1))


@parametrized_connection
@pytest.mark.parametrize("callback_as_partial", [True, False])
@validate_code_level_metrics(
    "test_pika_async_connection_consume" + (".test_async_connection_basic_get_inside_txn.<locals>" if six.PY3 else ""),
    "on_message",
)
@validate_transaction_metrics(
    ("test_pika_async_connection_consume:" "test_async_connection_basic_get_inside_txn"),
    scoped_metrics=_test_select_conn_basic_get_inside_txn_metrics,
    rollup_metrics=_test_select_conn_basic_get_inside_txn_metrics,
    background_task=True,
)
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
@background_task()
def test_async_connection_basic_get_inside_txn(producer, ConnectionClass, callback_as_partial):
    def on_message(channel, method_frame, header_frame, body):
        assert method_frame
        assert body == BODY
        channel.basic_ack(method_frame.delivery_tag)
        channel.close()
        connection.close()
        connection.ioloop.stop()

    if callback_as_partial:
        on_message = functools.partial(on_message)

    def on_open_channel(channel):
        channel.basic_get(callback=on_message, queue=QUEUE)

    def on_open_connection(connection):
        connection.channel(on_open_callback=on_open_channel)

    connection = ConnectionClass(pika.ConnectionParameters(DB_SETTINGS["host"]), on_open_callback=on_open_connection)

    try:
        connection.ioloop.start()
    except:
        connection.close()
        connection.ioloop.stop()
        raise


@parametrized_connection
@pytest.mark.parametrize("callback_as_partial", [True, False])
def test_select_connection_basic_get_outside_txn(producer, ConnectionClass, callback_as_partial):
    metrics_list = []

    @capture_transaction_metrics(metrics_list)
    def test_basic_get():
        def on_message(channel, method_frame, header_frame, body):
            assert method_frame
            assert body == BODY
            channel.basic_ack(method_frame.delivery_tag)
            channel.close()
            connection.close()
            connection.ioloop.stop()

        if callback_as_partial:
            on_message = functools.partial(on_message)

        def on_open_channel(channel):
            channel.basic_get(callback=on_message, queue=QUEUE)

        def on_open_connection(connection):
            connection.channel(on_open_callback=on_open_channel)

        connection = ConnectionClass(
            pika.ConnectionParameters(DB_SETTINGS["host"]), on_open_callback=on_open_connection
        )

        try:
            connection.ioloop.start()
        except:
            connection.close()
            connection.ioloop.stop()
            raise

    test_basic_get()

    # Confirm that no metrics have been created. This is because no background
    # task should be created for basic_get actions.
    assert not metrics_list


_test_select_conn_basic_get_inside_txn_no_callback_metrics = [
    ("MessageBroker/RabbitMQ/Exchange/Produce/Named/%s" % EXCHANGE, None),
    ("MessageBroker/RabbitMQ/Exchange/Consume/Named/%s" % EXCHANGE, None),
]


@pytest.mark.skipif(
    condition=pika_version_info[0] > 0, reason="pika 1.0 removed the ability to use basic_get with callback=None"
)
@parametrized_connection
@validate_transaction_metrics(
    ("test_pika_async_connection_consume:" "test_async_connection_basic_get_inside_txn_no_callback"),
    scoped_metrics=_test_select_conn_basic_get_inside_txn_no_callback_metrics,
    rollup_metrics=_test_select_conn_basic_get_inside_txn_no_callback_metrics,
    background_task=True,
)
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
@background_task()
def test_async_connection_basic_get_inside_txn_no_callback(producer, ConnectionClass):
    def on_open_channel(channel):
        channel.basic_get(callback=None, queue=QUEUE)
        channel.close()
        connection.close()
        connection.ioloop.stop()

    def on_open_connection(connection):
        connection.channel(on_open_callback=on_open_channel)

    connection = ConnectionClass(pika.ConnectionParameters(DB_SETTINGS["host"]), on_open_callback=on_open_connection)

    try:
        connection.ioloop.start()
    except:
        connection.close()
        connection.ioloop.stop()
        raise


_test_async_connection_basic_get_empty_metrics = [
    ("MessageBroker/RabbitMQ/Exchange/Produce/Named/%s" % EXCHANGE, None),
    ("MessageBroker/RabbitMQ/Exchange/Consume/Named/%s" % EXCHANGE, None),
]


@parametrized_connection
@pytest.mark.parametrize("callback_as_partial", [True, False])
@validate_transaction_metrics(
    ("test_pika_async_connection_consume:" "test_async_connection_basic_get_empty"),
    scoped_metrics=_test_async_connection_basic_get_empty_metrics,
    rollup_metrics=_test_async_connection_basic_get_empty_metrics,
    background_task=True,
)
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
@background_task()
def test_async_connection_basic_get_empty(ConnectionClass, callback_as_partial):
    QUEUE = "test_async_empty"

    def on_message(channel, method_frame, header_frame, body):
        assert False, body.decode("UTF-8")

    if callback_as_partial:
        on_message = functools.partial(on_message)

    def on_open_channel(channel):
        channel.basic_get(callback=on_message, queue=QUEUE)
        channel.close()
        connection.close()
        connection.ioloop.stop()

    def on_open_connection(connection):
        connection.channel(on_open_callback=on_open_channel)

    connection = ConnectionClass(pika.ConnectionParameters(DB_SETTINGS["host"]), on_open_callback=on_open_connection)

    try:
        connection.ioloop.start()
    except:
        connection.close()
        connection.ioloop.stop()
        raise


_test_select_conn_basic_consume_in_txn_metrics = [
    ("MessageBroker/RabbitMQ/Exchange/Produce/Named/%s" % EXCHANGE, None),
    ("MessageBroker/RabbitMQ/Exchange/Consume/Named/%s" % EXCHANGE, None),
]

if six.PY3:
    _test_select_conn_basic_consume_in_txn_metrics.append(
        (
            (
                "Function/test_pika_async_connection_consume:"
                "test_async_connection_basic_consume_inside_txn."
                "<locals>.on_message"
            ),
            1,
        )
    )
else:
    _test_select_conn_basic_consume_in_txn_metrics.append(("Function/test_pika_async_connection_consume:on_message", 1))


@parametrized_connection
@validate_transaction_metrics(
    ("test_pika_async_connection_consume:" "test_async_connection_basic_consume_inside_txn"),
    scoped_metrics=_test_select_conn_basic_consume_in_txn_metrics,
    rollup_metrics=_test_select_conn_basic_consume_in_txn_metrics,
    background_task=True,
)
@validate_code_level_metrics(
    "test_pika_async_connection_consume"
    + (".test_async_connection_basic_consume_inside_txn.<locals>" if six.PY3 else ""),
    "on_message",
)
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
@background_task()
def test_async_connection_basic_consume_inside_txn(producer, ConnectionClass):
    def on_message(channel, method_frame, header_frame, body):
        assert hasattr(method_frame, "_nr_start_time")
        assert body == BODY
        channel.basic_ack(method_frame.delivery_tag)
        channel.close()
        connection.close()
        connection.ioloop.stop()

    def on_open_channel(channel):
        basic_consume(channel, QUEUE, on_message)

    def on_open_connection(connection):
        connection.channel(on_open_callback=on_open_channel)

    connection = ConnectionClass(pika.ConnectionParameters(DB_SETTINGS["host"]), on_open_callback=on_open_connection)

    try:
        connection.ioloop.start()
    except:
        connection.close()
        connection.ioloop.stop()
        raise


_test_select_conn_basic_consume_two_exchanges = [
    ("MessageBroker/RabbitMQ/Exchange/Produce/Named/%s" % EXCHANGE, None),
    ("MessageBroker/RabbitMQ/Exchange/Consume/Named/%s" % EXCHANGE, None),
    ("MessageBroker/RabbitMQ/Exchange/Produce/Named/%s" % EXCHANGE_2, None),
    ("MessageBroker/RabbitMQ/Exchange/Consume/Named/%s" % EXCHANGE_2, None),
]

if six.PY3:
    _test_select_conn_basic_consume_two_exchanges.append(
        (
            (
                "Function/test_pika_async_connection_consume:"
                "test_async_connection_basic_consume_two_exchanges."
                "<locals>.on_message_1"
            ),
            1,
        )
    )
    _test_select_conn_basic_consume_two_exchanges.append(
        (
            (
                "Function/test_pika_async_connection_consume:"
                "test_async_connection_basic_consume_two_exchanges."
                "<locals>.on_message_2"
            ),
            1,
        )
    )
else:
    _test_select_conn_basic_consume_two_exchanges.append(
        ("Function/test_pika_async_connection_consume:on_message_1", 1)
    )
    _test_select_conn_basic_consume_two_exchanges.append(
        ("Function/test_pika_async_connection_consume:on_message_2", 1)
    )


@parametrized_connection
@validate_transaction_metrics(
    ("test_pika_async_connection_consume:" "test_async_connection_basic_consume_two_exchanges"),
    scoped_metrics=_test_select_conn_basic_consume_two_exchanges,
    rollup_metrics=_test_select_conn_basic_consume_two_exchanges,
    background_task=True,
)
@validate_code_level_metrics(
    "test_pika_async_connection_consume"
    + (".test_async_connection_basic_consume_two_exchanges.<locals>" if six.PY3 else ""),
    "on_message_1",
)
@validate_code_level_metrics(
    "test_pika_async_connection_consume"
    + (".test_async_connection_basic_consume_two_exchanges.<locals>" if six.PY3 else ""),
    "on_message_2",
)
@background_task()
def test_async_connection_basic_consume_two_exchanges(producer, producer_2, ConnectionClass):
    global events_received
    events_received = 0

    def on_message_1(channel, method_frame, header_frame, body):
        channel.basic_ack(method_frame.delivery_tag)
        assert hasattr(method_frame, "_nr_start_time")
        assert body == BODY

        global events_received
        events_received += 1

        if events_received == 2:
            channel.close()
            connection.close()
            connection.ioloop.stop()

    def on_message_2(channel, method_frame, header_frame, body):
        channel.basic_ack(method_frame.delivery_tag)
        assert hasattr(method_frame, "_nr_start_time")
        assert body == BODY

        global events_received
        events_received += 1

        if events_received == 2:
            channel.close()
            connection.close()
            connection.ioloop.stop()

    def on_open_channel(channel):
        basic_consume(channel, QUEUE, on_message_1)
        basic_consume(channel, QUEUE_2, on_message_2)

    def on_open_connection(connection):
        connection.channel(on_open_callback=on_open_channel)

    connection = ConnectionClass(pika.ConnectionParameters(DB_SETTINGS["host"]), on_open_callback=on_open_connection)

    try:
        connection.ioloop.start()
    except:
        connection.close()
        connection.ioloop.stop()
        raise


# This should not create a transaction
@function_not_called("newrelic.core.stats_engine", "StatsEngine.record_transaction")
@override_application_settings({"debug.record_transaction_failure": True})
def test_tornado_connection_basic_consume_outside_transaction(producer):
    def on_message(channel, method_frame, header_frame, body):
        assert hasattr(method_frame, "_nr_start_time")
        assert body == BODY
        channel.basic_ack(method_frame.delivery_tag)
        channel.close()
        connection.close()
        connection.ioloop.stop()

    def on_open_channel(channel):
        basic_consume(channel, QUEUE, on_message)

    def on_open_connection(connection):
        connection.channel(on_open_callback=on_open_channel)

    connection = TornadoConnection(pika.ConnectionParameters(DB_SETTINGS["host"]), on_open_callback=on_open_connection)

    try:
        connection.ioloop.start()
    except:
        connection.close()
        connection.ioloop.stop()
        raise


if six.PY3:
    _txn_name = (
        "test_pika_async_connection_consume:"
        "test_select_connection_basic_consume_outside_transaction."
        "<locals>.on_message"
    )
    _test_select_connection_consume_outside_txn_metrics = [
        (
            (
                "Function/test_pika_async_connection_consume:"
                "test_select_connection_basic_consume_outside_transaction."
                "<locals>.on_message"
            ),
            None,
        )
    ]
else:
    _txn_name = "test_pika_async_connection_consume:on_message"
    _test_select_connection_consume_outside_txn_metrics = [
        ("Function/test_pika_async_connection_consume:on_message", None)
    ]


# This should create a transaction
@validate_transaction_metrics(
    _txn_name,
    scoped_metrics=_test_select_connection_consume_outside_txn_metrics,
    rollup_metrics=_test_select_connection_consume_outside_txn_metrics,
    background_task=True,
    group="Message/RabbitMQ/Exchange/%s" % EXCHANGE,
)
@validate_code_level_metrics(
    "test_pika_async_connection_consume"
    + (".test_select_connection_basic_consume_outside_transaction.<locals>" if six.PY3 else ""),
    "on_message",
)
def test_select_connection_basic_consume_outside_transaction(producer):
    def on_message(channel, method_frame, header_frame, body):
        assert hasattr(method_frame, "_nr_start_time")
        assert body == BODY
        channel.basic_ack(method_frame.delivery_tag)
        channel.close()
        connection.close()
        connection.ioloop.stop()

    def on_open_channel(channel):
        basic_consume(channel, QUEUE, on_message)

    def on_open_connection(connection):
        connection.channel(on_open_callback=on_open_channel)

    connection = pika.SelectConnection(
        pika.ConnectionParameters(DB_SETTINGS["host"]), on_open_callback=on_open_connection
    )

    try:
        connection.ioloop.start()
    except:
        connection.close()
        connection.ioloop.stop()
        raise



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from newrelic.core.application import Application
from newrelic.core.config import global_settings
from newrelic.network.exceptions import ForceAgentDisconnect

from testing_support.fixtures import (
    override_generic_settings,
    failing_endpoint
)
from testing_support.validators.validate_internal_metrics import validate_internal_metrics


SETTINGS = global_settings()


@override_generic_settings(SETTINGS, {
    'developer_mode': True,
})
@failing_endpoint('preconnect', raises=ForceAgentDisconnect)
def test_http_gone_stops_connect():
    app = Application('Python Agent Test (agent_unittests-connect)')
    app.connect_to_data_collector(None)

    # The agent must not reattempt a connection after a ForceAgentDisconnect.
    # If it does, we'll end up with a session here.
    assert not app._active_session


_logging_settings_matrix = [
    (True, True),
    (True, False),
    (False, True),
    (False, False),
]


@override_generic_settings(SETTINGS, {
    'developer_mode': True,
})
@pytest.mark.parametrize("feature_setting,subfeature_setting", _logging_settings_matrix)
def test_logging_connect_supportability_metrics(feature_setting, subfeature_setting):
    metric_value = "enabled" if feature_setting and subfeature_setting else "disabled"

    @override_generic_settings(SETTINGS, {
        "application_logging.enabled": feature_setting,
        "application_logging.forwarding.enabled": subfeature_setting,
        "application_logging.metrics.enabled": subfeature_setting,
        "application_logging.local_decorating.enabled": subfeature_setting,
    })
    @validate_internal_metrics([
        ("Supportability/Logging/Forwarding/Python/%s" % metric_value, 1),
        ("Supportability/Logging/LocalDecorating/Python/%s" % metric_value, 1),
        ("Supportability/Logging/Metrics/Python/%s" % metric_value, 1),
    ])
    def test():
        app = Application('Python Agent Test (agent_unittests-connect)')
        app.connect_to_data_collector(None)

        assert app._active_session
    
    test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import newrelic.packages.urllib3 as urllib3

try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode

from newrelic.common.agent_http import BaseClient


def create_client_cls(status, data, url=None):
    if url:
        expected = urllib3.util.parse_url(url)
        if expected.port:
            expected_port = expected.port
        elif expected.scheme == 'https':
            expected_port = 443
        elif expected.scheme == 'http':
            expected_port = 80
        else:
            assert False, "Expected URL is missing scheme and port"
    else:
        expected = None

    class MockHttpClient(BaseClient):
        FAIL = False

        def __init__(self, host, port=80, *args, **kwargs):
            self.host = host
            self.port = port

        def send_request(
                self,
                method="POST",
                path="/agent_listener/invoke_raw_method",
                params=None,
                headers=None,
                payload=None,
        ):
            if expected:
                try:
                    assert self.host == expected.host
                    assert self.port == expected_port
                    assert path == expected.path
                    if expected.query:
                        query = urlencode(params)
                        assert query == expected.query
                except:
                    MockHttpClient.FAIL = True

            if status == 0:
                raise Exception

            return status, data

    return MockHttpClient



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from conftest import cache_kafka_consumer_headers
from testing_support.fixtures import (
    reset_core_stats_engine,
    validate_attributes,
    validate_error_event_attributes_outside_transaction,
)
from testing_support.validators.validate_distributed_trace_accepted import (
    validate_distributed_trace_accepted,
)
from testing_support.validators.validate_transaction_count import (
    validate_transaction_count,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.api.transaction import end_of_transaction
from newrelic.common.object_names import callable_name
from newrelic.packages import six


def test_custom_metrics(get_consumer_record, topic):
    @validate_transaction_metrics(
        "Named/%s" % topic,
        group="Message/Kafka/Topic",
        custom_metrics=[
            ("Message/Kafka/Topic/Named/%s/Received/Bytes" % topic, 1),
            ("Message/Kafka/Topic/Named/%s/Received/Messages" % topic, 1),
        ],
        background_task=True,
    )
    def _test():
        get_consumer_record()

    _test()


def test_multiple_transactions(get_consumer_record, topic):
    @validate_transaction_count(2)
    def _test():
        get_consumer_record()
        get_consumer_record()

    _test()


def test_custom_metrics_on_existing_transaction(get_consumer_record, topic):
    from kafka.version import __version__ as version

    transaction_name = (
        "test_consumer:test_custom_metrics_on_existing_transaction.<locals>._test" if six.PY3 else "test_consumer:_test"
    )

    @validate_transaction_metrics(
        transaction_name,
        custom_metrics=[
            ("Message/Kafka/Topic/Named/%s/Received/Bytes" % topic, 1),
            ("Message/Kafka/Topic/Named/%s/Received/Messages" % topic, 1),
            ("Python/MessageBroker/Kafka-Python/%s" % version, 1),
        ],
        background_task=True,
    )
    @validate_transaction_count(1)
    @background_task()
    def _test():
        get_consumer_record()

    _test()


def test_custom_metrics_inactive_transaction(get_consumer_record, topic):
    transaction_name = (
        "test_consumer:test_custom_metrics_inactive_transaction.<locals>._test" if six.PY3 else "test_consumer:_test"
    )

    @validate_transaction_metrics(
        transaction_name,
        custom_metrics=[
            ("Message/Kafka/Topic/Named/%s/Received/Bytes" % topic, None),
            ("Message/Kafka/Topic/Named/%s/Received/Messages" % topic, None),
        ],
        background_task=True,
    )
    @validate_transaction_count(1)
    @background_task()
    def _test():
        end_of_transaction()
        get_consumer_record()

    _test()


def test_agent_attributes(get_consumer_record):
    @validate_attributes("agent", ["kafka.consume.client_id", "kafka.consume.byteCount"])
    def _test():
        get_consumer_record()

    _test()


def test_consumer_errors(get_consumer_record, consumer_next_raises):
    exc_class = RuntimeError

    @reset_core_stats_engine()
    @validate_error_event_attributes_outside_transaction(
        num_errors=1, exact_attrs={"intrinsic": {"error.class": callable_name(exc_class)}, "agent": {}, "user": {}}
    )
    def _test():
        with pytest.raises(exc_class):
            get_consumer_record()

    _test()


def test_consumer_handled_errors_not_recorded(get_consumer_record):
    # It's important to check that we do not notice the StopIteration error.
    @validate_transaction_errors([])
    def _test():
        get_consumer_record()

    _test()


def test_distributed_tracing_headers(topic, producer, consumer, serialize):
    # Produce the messages inside a transaction, making sure to close it.
    @background_task()
    def _produce():
        producer.send(topic, key=serialize("bar"), value=serialize({"foo": 1}))
        producer.flush()

    @validate_transaction_metrics(
        "Named/%s" % topic,
        group="Message/Kafka/Topic",
        rollup_metrics=[
            ("Supportability/DistributedTrace/AcceptPayload/Success", None),
            ("Supportability/TraceContext/Accept/Success", 1),
        ],
        background_task=True,
    )
    @validate_transaction_count(1)
    def _consume():
        consumer_iter = iter(consumer)

        @validate_distributed_trace_accepted(transport_type="Kafka")
        @cache_kafka_consumer_headers
        def _test():
            # Start the transaction but don't exit it.
            timeout = 10
            attempts = 0
            record = None
            while not record and attempts < timeout:
                try:
                    record = next(consumer_iter)
                except StopIteration:
                    attempts += 1

        _test()

        # Exit the transaction.
        with pytest.raises(StopIteration):
            next(consumer_iter)

    _produce()
    _consume()


@pytest.fixture()
def consumer_next_raises(consumer):
    def _poll(*args, **kwargs):
        raise RuntimeError()

    consumer.poll = _poll
    return consumer



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""These tests check two things:

    1. Using a notfound_view will name the transaction after the not_found
       view, not the default_exceptionresponse_view. There can only be one
       notfound_view per Pyramid application, so to test both cases (default
       and user-defined notfound_view), we use 2 separate apps.

    2. Verifying that we don't double wrap the notfound_view. Because
       we instrument Configurator._derive_view(), and because
       Configurator._derive_view() is called twice when adding a
       notfound_view (once in add_notfound_view(), and again in add_view()),
       our instrumentation checks to see if a view has already been
       wrapped before wrapping.

       If our check fails, then the metric '_test_append_slash_app:not_found'
       will have a count of 2.
"""

import pkg_resources
import pytest
import re

from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_transaction_errors import validate_transaction_errors

def _to_int(version_str):
    m = re.match(r'\d+', version_str)
    return int(m.group(0)) if m else 0

def pyramid_version():
    s = pkg_resources.get_distribution('pyramid').version
    return tuple([_to_int(c) for c in s.split('.')[:2]])

# Defining a `pytestmark` at the module level means py.test will apply
# this `skipif` conditional to all tests in the module.

pytestmark = pytest.mark.skipif(pyramid_version() < (1, 3),
        reason='requires pyramid >= (1, 3)')

def target_application():
    # We need to delay Pyramid application creation because of ordering
    # issues whereby the agent needs to be initialised before Pyramid is
    # imported and the routes configured. Normally pytest only runs the
    # global fixture which will initialise the agent after each test
    # file is imported, which is too late. We also can't do application
    # creation within a function as Pyramid relies on view handlers being
    # at global scope, so import it from a separate module.

    from _test_append_slash_app import _test_application
    return _test_application

_test_append_slash_app_index_scoped_metrics = [
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/pyramid.router:Router.__call__', 1),
        ('Function/_test_append_slash_app:home_view', 1)]

@validate_transaction_errors(errors=[])
@validate_transaction_metrics('_test_append_slash_app:home_view',
        scoped_metrics=_test_append_slash_app_index_scoped_metrics)
def test_index():
    application = target_application()
    response = application.get('/')
    response.mustcontain('INDEX RESPONSE')

_test_not_found_append_slash_scoped_metrics = [
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/pyramid.router:Router.__call__', 1),
        ('Function/pyramid.view:AppendSlashNotFoundViewFactory', 1),
        ('Function/_test_append_slash_app:not_found', 1)]

@validate_transaction_errors(errors=[])
@validate_transaction_metrics('_test_append_slash_app:not_found',
        scoped_metrics=_test_not_found_append_slash_scoped_metrics)
def test_not_found_append_slash():
    application = target_application()
    response = application.get('/foo', status=404)
    response.mustcontain('NOT FOUND')



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module provides functions to collect information about the operating
system, Python and hosting environment.

"""

import os
import platform
import sys
import sysconfig

import newrelic
from newrelic.common.package_version_utils import get_package_version
from newrelic.common.system_info import (
    logical_processor_count,
    physical_processor_count,
    total_physical_memory,
)

try:
    import newrelic.core._thread_utilization
except ImportError:
    pass


def environment_settings():
    """Returns an array of arrays of environment settings"""
    env = []

    # Agent information.

    env.append(("Agent Version", ".".join(map(str, newrelic.version_info))))

    # System information.

    env.append(("Arch", platform.machine()))
    env.append(("OS", platform.system()))
    env.append(("OS version", platform.release()))

    env.append(("Total Physical Memory (MB)", total_physical_memory()))
    env.append(("Logical Processors", logical_processor_count()))

    physical_processor_packages, physical_cores = physical_processor_count()

    # Report this attribute only if it has a valid value.

    if physical_processor_packages:
        env.append(("Physical Processor Packages", physical_processor_packages))

    # Report this attribute only if it has a valid value.

    if physical_cores:
        env.append(("Physical Cores", physical_cores))

    # Python information.

    env.append(("Python Program Name", sys.argv[0]))

    env.append(("Python Executable", sys.executable))

    env.append(("Python Home", os.environ.get("PYTHONHOME", "")))
    env.append(("Python Path", os.environ.get("PYTHONPATH", "")))

    env.append(("Python Prefix", sys.prefix))
    env.append(("Python Exec Prefix", sys.exec_prefix))

    env.append(("Python Runtime", ".".join(platform.python_version_tuple())))

    env.append(("Python Implementation", platform.python_implementation()))
    env.append(("Python Version", sys.version))

    env.append(("Python Platform", sys.platform))

    env.append(("Python Max Unicode", sys.maxunicode))

    # Extensions information.

    extensions = []

    if "newrelic.core._thread_utilization" in sys.modules:
        extensions.append("newrelic.core._thread_utilization")

    env.append(("Compiled Extensions", ", ".join(extensions)))

    # Dispatcher information.

    dispatcher = []

    # Find the first dispatcher module that's been loaded and report that as the dispatcher.
    # If possible, also report the dispatcher's version and any other environment information.
    if not dispatcher and "mod_wsgi" in sys.modules:
        mod_wsgi = sys.modules["mod_wsgi"]
        if hasattr(mod_wsgi, "process_group"):
            if mod_wsgi.process_group == "":
                dispatcher.append(("Dispatcher", "Apache/mod_wsgi (embedded)"))
            else:
                dispatcher.append(("Dispatcher", "Apache/mod_wsgi (daemon)"))
            env.append(("Apache/mod_wsgi Process Group", mod_wsgi.process_group))
        else:
            dispatcher.append(("Dispatcher", "Apache/mod_wsgi"))
        if hasattr(mod_wsgi, "version"):
            dispatcher.append(("Dispatcher Version", str(mod_wsgi.version)))
        if hasattr(mod_wsgi, "application_group"):
            env.append(("Apache/mod_wsgi Application Group", mod_wsgi.application_group))

    if not dispatcher and "uwsgi" in sys.modules:
        dispatcher.append(("Dispatcher", "uWSGI"))
        uwsgi = sys.modules["uwsgi"]
        if hasattr(uwsgi, "version"):
            dispatcher.append(("Dispatcher Version", uwsgi.version))

    if not dispatcher and "flup.server.fcgi" in sys.modules:
        dispatcher.append(("Dispatcher", "flup/fastcgi (threaded)"))

    if not dispatcher and "flup.server.fcgi_fork" in sys.modules:
        dispatcher.append(("Dispatcher", "flup/fastcgi (prefork)"))

    if not dispatcher and "flup.server.scgi" in sys.modules:
        dispatcher.append(("Dispatcher", "flup/scgi (threaded)"))

    if not dispatcher and "flup.server.scgi_fork" in sys.modules:
        dispatcher.append(("Dispatcher", "flup/scgi (prefork)"))

    if not dispatcher and "flup.server.ajp" in sys.modules:
        dispatcher.append(("Dispatcher", "flup/ajp (threaded)"))

    if not dispatcher and "flup.server.ajp_fork" in sys.modules:
        dispatcher.append(("Dispatcher", "flup/ajp (forking)"))

    if not dispatcher and "flup.server.cgi" in sys.modules:
        dispatcher.append(("Dispatcher", "flup/cgi"))

    if not dispatcher and "gunicorn" in sys.modules:
        if "gunicorn.workers.ggevent" in sys.modules:
            dispatcher.append(("Dispatcher", "gunicorn (gevent)"))
        elif "gunicorn.workers.geventlet" in sys.modules:
            dispatcher.append(("Dispatcher", "gunicorn (eventlet)"))
        elif "uvicorn.workers" in sys.modules:
            dispatcher.append(("Dispatcher", "gunicorn (uvicorn)"))
            uvicorn = sys.modules.get("uvicorn")
            if hasattr(uvicorn, "__version__"):
                dispatcher.append(("Worker Version", uvicorn.__version__))
        else:
            dispatcher.append(("Dispatcher", "gunicorn"))

        gunicorn = sys.modules["gunicorn"]
        if hasattr(gunicorn, "__version__"):
            dispatcher.append(("Dispatcher Version", gunicorn.__version__))

    if not dispatcher and "uvicorn" in sys.modules:
        dispatcher.append(("Dispatcher", "uvicorn"))
        uvicorn = sys.modules["uvicorn"]

        if hasattr(uvicorn, "__version__"):
            dispatcher.append(("Dispatcher Version", uvicorn.__version__))

    if not dispatcher and "hypercorn" in sys.modules:
        dispatcher.append(("Dispatcher", "hypercorn"))
        hypercorn = sys.modules["hypercorn"]

        if hasattr(hypercorn, "__version__"):
            dispatcher.append(("Dispatcher Version", hypercorn.__version__))
        else:
            try:
                dispatcher.append(("Dispatcher Version", get_package_version("hypercorn")))
            except Exception:
                pass

    if not dispatcher and "daphne" in sys.modules:
        dispatcher.append(("Dispatcher", "daphne"))
        daphne = sys.modules["daphne"]

        if hasattr(daphne, "__version__"):
            dispatcher.append(("Dispatcher Version", daphne.__version__))

    if not dispatcher and "tornado" in sys.modules:
        dispatcher.append(("Dispatcher", "tornado"))
        tornado = sys.modules["tornado"]
        if hasattr(tornado, "version_info"):
            dispatcher.append(("Dispatcher Version", str(tornado.version_info)))

    env.extend(dispatcher)

    # Module information.
    purelib = sysconfig.get_path("purelib")
    platlib = sysconfig.get_path("platlib")

    plugins = []

    # Using any iterable to create a snapshot of sys.modules can occassionally
    # fail in a rare case when modules are imported in parallel by different
    # threads.
    #
    # TL;DR: Do NOT use an iterable on the original sys.modules to generate the
    # list
    for name, module in sys.modules.copy().items():
        # Exclude lib.sub_paths as independent modules except for newrelic.hooks.
        if "." in name and not name.startswith("newrelic.hooks."):
            continue
        # If the module isn't actually loaded (such as failed relative imports
        # in Python 2.7), the module will be None and should not be reported.
        if not module:
            continue
        # Exclude standard library/built-in modules.
        # Third-party modules can be installed in either purelib or platlib directories.
        # See https://docs.python.org/3/library/sysconfig.html#installation-paths.
        if (
            not hasattr(module, "__file__")
            or not module.__file__
            or not module.__file__.startswith(purelib)
            or not module.__file__.startswith(platlib)
        ):
            continue

        try:
            version = get_package_version(name)
            plugins.append("%s (%s)" % (name, version))
        except Exception:
            plugins.append(name)

    env.append(("Plugin List", plugins))

    return env



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
import time

from testing_support.util import get_open_port

from gevent import Timeout, sleep

def request_timeout_application(environ, start_response):
    with Timeout(0.1):
        sleep(1.0)

    status = '200 OK'

    response_headers = [('Content-type', 'text/plain')]
    start_response(status, response_headers)

    try:
        yield b'WSGI RESPONSE'
    finally:
        pass

def request_timeout_response(environ, start_response):
    status = '200 OK'

    response_headers = [('Content-type', 'text/plain')]
    start_response(status, response_headers)

    yield b'WSGI'

    with Timeout(0.1):
        sleep(1.0)

    yield b' '
    yield b'RESPONSE'

def request_timeout_finalize(environ, start_response):
    status = '200 OK'

    response_headers = [('Content-type', 'text/plain')]
    start_response(status, response_headers)

    try:
        yield b'WSGI RESPONSE'

    finally:
        with Timeout(0.1):
            sleep(1.0)

def raise_exception_application(environ, start_response):
    raise RuntimeError('raise_exception_application')

    status = '200 OK'
    output = b'WSGI RESPONSE'

    response_headers = [('Content-type', 'text/plain'),
                        ('Content-Length', str(len(output)))]
    start_response(status, response_headers)

    return [output]

def raise_exception_response(environ, start_response):
    status = '200 OK'

    response_headers = [('Content-type', 'text/plain')]
    start_response(status, response_headers)

    yield b'WSGI'

    raise RuntimeError('raise_exception_response')

    yield b' '
    yield b'RESPONSE'

def raise_exception_finalize(environ, start_response):
    status = '200 OK'

    response_headers = [('Content-type', 'text/plain')]
    start_response(status, response_headers)

    try:
        yield b'WSGI RESPONSE'

    finally:
        raise RuntimeError('raise_exception_finalize')

def application_index(environ, start_response):
    status = '200 OK'

    response_headers = [('Content-type', 'text/plain')]
    start_response(status, response_headers)

    time.sleep(0.1)

    try:
        yield b'WSGI RESPONSE'
    finally:
        pass

def sample_application(environ, start_response):
    path_info = environ.get('PATH_INFO')

    if path_info.startswith('/request-timeout-application'):
        return request_timeout_application(environ, start_response)
    elif path_info.startswith('/request-timeout-response'):
        return request_timeout_response(environ, start_response)
    elif path_info.startswith('/request-timeout-finalize'):
        return request_timeout_finalize(environ, start_response)
    elif path_info.startswith('/raise-exception-application'):
        return raise_exception_application(environ, start_response)
    elif path_info.startswith('/raise-exception-response'):
        return raise_exception_response(environ, start_response)
    elif path_info.startswith('/raise-exception-finalize'):
        return raise_exception_finalize(environ, start_response)

    return application_index(environ, start_response)

def setup_application():
    port = get_open_port()

    def run_wsgi():
        from gevent.pywsgi import WSGIServer
        WSGIServer(('127.0.0.1', port), sample_application).serve_forever()

    wsgi_thread = threading.Thread(target=run_wsgi)
    wsgi_thread.daemon = True
    wsgi_thread.start()

    time.sleep(1)

    return port



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools


def exercise_function():
    return


class ExerciseClass(object):
    def exercise_method(self):
        return

    @staticmethod
    def exercise_static_method():
        return

    @classmethod
    def exercise_class_method(cls):
        return


class ExerciseClassCallable(object):
    def __call__(self):
        return


def exercise_method(self):
    return


@staticmethod
def exercise_static_method():
    return


@classmethod
def exercise_class_method(cls):
    return


def __call__(self):
    return


type_dict = {
    "exercise_method": exercise_method,
    "exercise_static_method": exercise_static_method,
    "exercise_class_method": exercise_class_method,
    "exercise_lambda": lambda: None,
}
callable_type_dict = type_dict.copy()
callable_type_dict["__call__"] = __call__

ExerciseTypeConstructor = type("ExerciseTypeConstructor", (object,), type_dict)
ExerciseTypeConstructorCallable = type("ExerciseTypeConstructorCallable", (object,), callable_type_dict)


CLASS_INSTANCE = ExerciseClass()
CLASS_INSTANCE_CALLABLE = ExerciseClassCallable()
TYPE_CONSTRUCTOR_CLASS_INSTANCE = ExerciseTypeConstructor()
TYPE_CONSTRUCTOR_CALLABLE_CLASS_INSTANCE = ExerciseTypeConstructorCallable()

exercise_lambda = lambda: None  # noqa: E731
exercise_partial = functools.partial(exercise_function)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from conftest import aioredis

import pytest
from conftest import AIOREDIS_VERSION, loop  # noqa
from testing_support.db_settings import redis_settings
from testing_support.fixtures import override_application_settings
from testing_support.util import instance_hostname
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task

DB_SETTINGS = redis_settings()

_enable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": True,
}
_disable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": False,
}

_base_scoped_metrics = (
    ("Datastore/operation/Redis/get", 1),
    ("Datastore/operation/Redis/set", 1),
    ("Datastore/operation/Redis/client_list", 1),
)

_base_rollup_metrics = (
    ("Datastore/all", 3),
    ("Datastore/allOther", 3),
    ("Datastore/Redis/all", 3),
    ("Datastore/Redis/allOther", 3),
    ("Datastore/operation/Redis/get", 1),
    ("Datastore/operation/Redis/set", 1),
    ("Datastore/operation/Redis/client_list", 1),
)

_concurrent_scoped_metrics = [
    ("Datastore/operation/Redis/get", 2),
    ("Datastore/operation/Redis/set", 2),
]

_concurrent_rollup_metrics = [
    ("Datastore/all", 4),
    ("Datastore/allOther", 4),
    ("Datastore/Redis/all", 4),
    ("Datastore/Redis/allOther", 4),
    ("Datastore/operation/Redis/set", 2),
    ("Datastore/operation/Redis/get", 2),
]

_disable_scoped_metrics = list(_base_scoped_metrics)
_disable_rollup_metrics = list(_base_rollup_metrics)

_enable_scoped_metrics = list(_base_scoped_metrics)
_enable_rollup_metrics = list(_base_rollup_metrics)


if len(DB_SETTINGS) > 1:
    redis_instance_1 = DB_SETTINGS[0]
    redis_instance_2 = DB_SETTINGS[1]

    _host_1 = instance_hostname(redis_instance_1["host"])
    _port_1 = redis_instance_1["port"]

    _host_2 = instance_hostname(redis_instance_2["host"])
    _port_2 = redis_instance_2["port"]

    instance_metric_name_1 = "Datastore/instance/Redis/%s/%s" % (_host_1, _port_1)
    instance_metric_name_2 = "Datastore/instance/Redis/%s/%s" % (_host_2, _port_2)

    _enable_rollup_metrics.extend(
        [
            (instance_metric_name_1, 2),
            (instance_metric_name_2, 1),
        ]
    )

    _disable_rollup_metrics.extend(
        [
            (instance_metric_name_1, None),
            (instance_metric_name_2, None),
        ]
    )
    _concurrent_rollup_metrics.extend(
        [
            (instance_metric_name_1, 2),
            (instance_metric_name_2, 2),
        ]
    )


@pytest.fixture(params=("Redis", "StrictRedis"))
def client_set(request, loop):  # noqa
    if len(DB_SETTINGS) > 1:
        if AIOREDIS_VERSION >= (2, 0):
            if request.param == "Redis":
                return (
                    aioredis.Redis(host=DB_SETTINGS[0]["host"], port=DB_SETTINGS[0]["port"], db=0),
                    aioredis.Redis(host=DB_SETTINGS[1]["host"], port=DB_SETTINGS[1]["port"], db=0),
                )
            elif request.param == "StrictRedis":
                return (
                    aioredis.StrictRedis(host=DB_SETTINGS[0]["host"], port=DB_SETTINGS[0]["port"], db=0),
                    aioredis.StrictRedis(host=DB_SETTINGS[1]["host"], port=DB_SETTINGS[1]["port"], db=0),
                )
            else:
                raise NotImplementedError()
        else:
            if request.param == "Redis":
                return (
                    loop.run_until_complete(
                        aioredis.create_redis("redis://%s:%d" % (DB_SETTINGS[0]["host"], DB_SETTINGS[0]["port"]), db=0)
                    ),
                    loop.run_until_complete(
                        aioredis.create_redis("redis://%s:%d" % (DB_SETTINGS[1]["host"], DB_SETTINGS[1]["port"]), db=0)
                    ),
                )
            elif request.param == "StrictRedis":
                pytest.skip("StrictRedis not implemented.")
            else:
                raise NotImplementedError()


async def exercise_redis(client_1, client_2):
    await client_1.set("key", "value")
    await client_1.get("key")

    if hasattr(client_2, "execute_command"):
        await client_2.execute_command("CLIENT", "LIST", parse="LIST")
    else:
        await client_2.execute("CLIENT", "LIST")


@pytest.mark.skipif(len(DB_SETTINGS) < 2, reason="Env not configured with multiple databases")
@override_application_settings(_enable_instance_settings)
@validate_transaction_metrics(
    "test_multiple_dbs:test_multiple_datastores_enabled",
    scoped_metrics=_enable_scoped_metrics,
    rollup_metrics=_enable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_multiple_datastores_enabled(client_set, loop):  # noqa
    loop.run_until_complete(exercise_redis(client_set[0], client_set[1]))


@pytest.mark.skipif(len(DB_SETTINGS) < 2, reason="Env not configured with multiple databases")
@override_application_settings(_disable_instance_settings)
@validate_transaction_metrics(
    "test_multiple_dbs:test_multiple_datastores_disabled",
    scoped_metrics=_disable_scoped_metrics,
    rollup_metrics=_disable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_multiple_datastores_disabled(client_set, loop):  # noqa
    loop.run_until_complete(exercise_redis(client_set[0], client_set[1]))


@pytest.mark.skipif(len(DB_SETTINGS) < 2, reason="Env not configured with multiple databases")
@validate_transaction_metrics(
    "test_multiple_dbs:test_concurrent_calls",
    scoped_metrics=_concurrent_scoped_metrics,
    rollup_metrics=_concurrent_rollup_metrics,
    background_task=True,
)
@override_application_settings(_enable_instance_settings)
@background_task()
def test_concurrent_calls(client_set, loop):  # noqa
    # Concurrent calls made with original instrumenation taken from synchonous Redis
    # instrumentation had a bug where datastore info on concurrent calls to multiple instances
    # would result in all instances reporting as the host/port of the final call made.

    import asyncio

    async def exercise_concurrent():
        await asyncio.gather(*(client.set("key-%d" % i, i) for i, client in enumerate(client_set)))
        await asyncio.gather(*(client.get("key-%d" % i) for i, client in enumerate(client_set)))

    loop.run_until_complete(exercise_concurrent())



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from conftest import (  # pylint: disable=E0611
    async_handler_support,
    skip_if_not_async_handler_support,
)
from testing_support.validators.validate_code_level_metrics import (
    validate_code_level_metrics,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

scoped_metrics = [
    ("Function/flask.app:Flask.wsgi_app", 1),
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/flask.app:Flask.preprocess_request", 1),
    ("Function/flask.app:Flask.process_response", 1),
    ("Function/flask.app:Flask.do_teardown_request", 1),
    ("Function/werkzeug.wsgi:ClosingIterator.close", 1),
]


def target_application():
    # We need to delay Flask application creation because of ordering
    # issues whereby the agent needs to be initialised before Flask is
    # imported and the routes configured. Normally pytest only runs the
    # global fixture which will initialise the agent after each test
    # file is imported, which is too late. We also can't do application
    # creation within a function as we will then get view handler
    # functions are different between Python 2 and 3, with the latter
    # showing <local> scope in path.

    if not async_handler_support:
        from _test_views import _test_application
    else:
        from _test_views_async import _test_application
    return _test_application


@validate_code_level_metrics("_test_views.TestView", "dispatch_request")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics("_test_views:test_view", scoped_metrics=scoped_metrics)
def test_class_based_view():
    application = target_application()
    response = application.get("/view")
    response.mustcontain("VIEW RESPONSE")


@skip_if_not_async_handler_support
@validate_code_level_metrics("_test_views_async.TestAsyncView", "dispatch_request")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics("_test_views_async:test_async_view", scoped_metrics=scoped_metrics)
def test_class_based_async_view():
    application = target_application()
    response = application.get("/async_view")
    response.mustcontain("ASYNC VIEW RESPONSE")


@validate_code_level_metrics("_test_views.TestMethodView", "get")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics("_test_views:test_methodview", scoped_metrics=scoped_metrics)
def test_get_method_view():
    application = target_application()
    response = application.get("/methodview")
    response.mustcontain("METHODVIEW GET RESPONSE")


@validate_code_level_metrics("_test_views.TestMethodView", "post")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics("_test_views:test_methodview", scoped_metrics=scoped_metrics)
def test_post_method_view():
    application = target_application()
    response = application.post("/methodview")
    response.mustcontain("METHODVIEW POST RESPONSE")


@skip_if_not_async_handler_support
@validate_code_level_metrics("_test_views_async.TestAsyncMethodView", "get")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics("_test_views_async:test_async_methodview", scoped_metrics=scoped_metrics)
def test_get_method_async_view():
    application = target_application()
    response = application.get("/async_methodview")
    response.mustcontain("ASYNC METHODVIEW GET RESPONSE")



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import threading
import time
from urllib.request import HTTPError, urlopen

import pkg_resources
import pytest
from testing_support.fixtures import (
    override_application_settings,
    raise_background_exceptions,
    wait_for_background_threads,
)
from testing_support.sample_asgi_applications import (
    AppWithCall,
    AppWithCallRaw,
    simple_app_v2_raw,
)
from testing_support.util import get_open_port
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.transaction import ignore_transaction
from newrelic.common.object_names import callable_name

HYPERCORN_VERSION = tuple(int(v) for v in pkg_resources.get_distribution("hypercorn").version.split("."))
asgi_2_unsupported = HYPERCORN_VERSION >= (0, 14, 1)
wsgi_unsupported = HYPERCORN_VERSION < (0, 14, 1)


def wsgi_app(environ, start_response):
    path = environ["PATH_INFO"]

    if path == "/":
        start_response("200 OK", response_headers=[])
    elif path == "/ignored":
        ignore_transaction()
        start_response("200 OK", response_headers=[])
    elif path == "/exc":
        raise ValueError("whoopsies")

    return []


@pytest.fixture(
    params=(
        pytest.param(
            simple_app_v2_raw,
            marks=pytest.mark.skipif(asgi_2_unsupported, reason="ASGI2 unsupported"),
        ),
        AppWithCallRaw(),
        AppWithCall(),
        pytest.param(
            wsgi_app,
            marks=pytest.mark.skipif(wsgi_unsupported, reason="WSGI unsupported"),
        ),
    ),
    ids=("raw", "class_with_call", "class_with_call_double_wrapped", "wsgi"),
)
def app(request):
    return request.param


@pytest.fixture()
def port(loop, app):
    import hypercorn.asyncio
    import hypercorn.config

    port = get_open_port()
    shutdown = asyncio.Event()

    def server_run():
        async def shutdown_trigger():
            await shutdown.wait()
            return True

        config = hypercorn.config.Config.from_mapping(
            {
                "bind": ["127.0.0.1:%d" % port],
            }
        )

        try:
            loop.run_until_complete(hypercorn.asyncio.serve(app, config, shutdown_trigger=shutdown_trigger))
        except Exception:
            pass

    thread = threading.Thread(target=server_run, daemon=True)
    thread.start()
    wait_for_port(port)
    yield port

    shutdown.set()
    loop.call_soon_threadsafe(loop.stop)
    thread.join(timeout=10)

    if thread.is_alive():
        raise RuntimeError("Thread failed to exit in time.")


def wait_for_port(port, retries=10):
    status = None
    for _ in range(retries):
        try:
            status = urlopen("http://localhost:%d/ignored" % port, timeout=1).status  # nosec
            assert status == 200
            return
        except Exception as e:
            status = e

        time.sleep(1)

    raise RuntimeError("Failed to wait for port %d. Got status %s" % (port, status))


@override_application_settings({"transaction_name.naming_scheme": "framework"})
def test_hypercorn_200(port, app):
    hypercorn_version = pkg_resources.get_distribution("hypercorn").version

    @validate_transaction_metrics(
        callable_name(app),
        custom_metrics=[
            ("Python/Dispatcher/Hypercorn/%s" % hypercorn_version, 1),
        ],
    )
    @raise_background_exceptions()
    @wait_for_background_threads()
    def response():
        return urlopen("http://localhost:%d" % port, timeout=10)  # nosec

    assert response().status == 200


@override_application_settings({"transaction_name.naming_scheme": "framework"})
def test_hypercorn_500(port, app):
    @validate_transaction_errors(["builtins:ValueError"])
    @validate_transaction_metrics(callable_name(app))
    @raise_background_exceptions()
    @wait_for_background_threads()
    def _test():
        with pytest.raises(HTTPError):
            urlopen("http://localhost:%d/exc" % port)  # nosec

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from billiard import get_context
from billiard.pool import Worker

from testing_support.validators.validate_function_called import (
        validate_function_called)


class OnExit(Exception):
    pass


@validate_function_called('newrelic.core.agent', 'Agent.shutdown_agent')
def test_max_tasks_per_child():

    def on_exit(*args, **kwargs):
        raise OnExit()

    ctx = get_context()
    worker = Worker(ctx.SimpleQueue(), ctx.SimpleQueue(), None,
            maxtasks=1, on_exit=on_exit)

    with pytest.raises(OnExit):
        worker._do_exit(None, 0)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" The purpose of these tests is to confirm that using a non-standard
connection pool that does not have a `connection_kwargs` attribute
will not result in an error.
"""

from testing_support.db_settings import redis_settings
from testing_support.fixture.event_loop import event_loop as loop  # noqa
from testing_support.fixtures import override_application_settings
from testing_support.util import instance_hostname
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task

DB_SETTINGS = redis_settings()[0]


class FakeConnectionPool(object):
    """Connection Pool without connection_kwargs attribute."""

    def __init__(self, connection):
        self.connection = connection

    async def get_connection(self, name=None, *keys, **options):
        return self.connection

    async def release(self, connection):
        await self.connection.disconnect()

    async def execute(self, *args, **kwargs):
        return await self.connection.execute(*args, **kwargs)

    # def execute_command(self, *args, **kwargs):
    #     return self.connection.execute_command(*args, **kwargs)


# Settings

_enable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": True,
}
_disable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": False,
}


_base_scoped_metrics = (
    ("Datastore/operation/Redis/get", 1),
    ("Datastore/operation/Redis/set", 1),
    ("Datastore/operation/Redis/client_list", 1),
)

_base_rollup_metrics = (
    ("Datastore/all", 3),
    ("Datastore/allOther", 3),
    ("Datastore/Redis/all", 3),
    ("Datastore/Redis/allOther", 3),
    ("Datastore/operation/Redis/get", 1),
    ("Datastore/operation/Redis/set", 1),
    ("Datastore/operation/Redis/client_list", 1),
)


_disable_scoped_metrics = list(_base_scoped_metrics)
_disable_rollup_metrics = list(_base_rollup_metrics)

_enable_scoped_metrics = list(_base_scoped_metrics)
_enable_rollup_metrics = list(_base_rollup_metrics)

_host = instance_hostname(DB_SETTINGS["host"])
_port = DB_SETTINGS["port"]

_instance_metric_name = "Datastore/instance/Redis/%s/%s" % (_host, _port)

_enable_rollup_metrics.append((_instance_metric_name, 3))

_disable_rollup_metrics.append((_instance_metric_name, None))


# Operations
async def exercise_redis(client):
    await client.set("key", "value")
    await client.get("key")
    if hasattr(client, "execute_command"):
        await client.execute_command("CLIENT", "LIST", parse="LIST")
    else:
        await client.execute("CLIENT", "LIST")


@override_application_settings(_enable_instance_settings)
@validate_transaction_metrics(
    "test_custom_conn_pool:test_fake_conn_pool_enable_instance",
    scoped_metrics=_enable_scoped_metrics,
    rollup_metrics=_enable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_fake_conn_pool_enable_instance(client, loop, monkeypatch):  # noqa
    # Get a real connection
    conn = getattr(client, "_pool_or_conn", None)
    if conn is None:
        conn = loop.run_until_complete(client.connection_pool.get_connection("GET"))

    # Replace the original connection pool with one that doesn't
    # have the `connection_kwargs` attribute.

    fake_pool = FakeConnectionPool(conn)
    monkeypatch.setattr(client, "connection_pool", fake_pool, raising=False)
    monkeypatch.setattr(client, "_pool_or_conn", fake_pool, raising=False)
    assert not hasattr(client.connection_pool, "connection_kwargs")

    loop.run_until_complete(exercise_redis(client))


@override_application_settings(_disable_instance_settings)
@validate_transaction_metrics(
    "test_custom_conn_pool:test_fake_conn_pool_disable_instance",
    scoped_metrics=_disable_scoped_metrics,
    rollup_metrics=_disable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_fake_conn_pool_disable_instance(client, loop, monkeypatch):  # noqa
    # Get a real connection
    conn = getattr(client, "_pool_or_conn", None)
    if conn is None:
        conn = loop.run_until_complete(client.connection_pool.get_connection("GET"))

    # Replace the original connection pool with one that doesn't
    # have the `connection_kwargs` attribute.

    fake_pool = FakeConnectionPool(conn)
    monkeypatch.setattr(client, "connection_pool", fake_pool, raising=False)
    monkeypatch.setattr(client, "_pool_or_conn", fake_pool, raising=False)
    assert not hasattr(client.connection_pool, "connection_kwargs")

    loop.run_until_complete(exercise_redis(client))



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from newrelic.api.transaction import current_transaction
from newrelic.common.encoding_utils import DistributedTracePayload

OUTBOUND_TRACE_KEYS_REQUIRED = (
        'ty', 'ac', 'ap', 'tr', 'pr', 'sa', 'ti')


def validate_distributed_tracing_header(header='newrelic'):
    transaction = current_transaction()
    headers = transaction._test_request_headers
    account_id = transaction.settings.account_id
    application_id = transaction.settings.primary_application_id

    assert header in headers, headers

    values = headers[header]
    if isinstance(values, list):
        assert len(values) == 1, headers
        assert isinstance(values[0], type(''))
        value = values[0]
    else:
        value = values

    # Parse payload
    payload = DistributedTracePayload.from_http_safe(value)

    # Distributed Tracing v0.1 is currently implemented
    assert payload['v'] == [0, 1], payload['v']

    data = payload['d']

    # Verify all required keys are present
    assert all(k in data for k in OUTBOUND_TRACE_KEYS_REQUIRED)

    # Type will always be App (not mobile / browser)
    assert data['ty'] == 'App'

    # Verify account/app id
    assert data['ac'] == account_id
    assert data['ap'] == application_id

    # Verify data belonging to this transaction
    assert data['tx'] == transaction.guid

    # If span events are enabled, id should be sent
    # otherwise, id should be omitted
    if transaction.settings.span_events.enabled:
        assert 'id' in data
    else:
        assert 'id' not in data

    # Verify referring transaction information
    assert len(data['tr']) == 32
    if transaction.referring_transaction_guid is not None:
        assert data['tr'] == transaction._trace_id
    else:
        assert data['tr'].startswith(transaction.guid)

    assert 'pa' not in data

    # Verify timestamp is an integer
    assert isinstance(data['ti'], int)

    # Verify that priority is a float
    assert isinstance(data['pr'], float)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Union

import strawberry.mutation
import strawberry.type
from strawberry import Schema, field
from strawberry.asgi import GraphQL
from strawberry.schema.config import StrawberryConfig
from strawberry.types.types import Optional


@strawberry.type
class Author:
    first_name: str
    last_name: str


@strawberry.type
class Book:
    id: int
    name: str
    isbn: str
    author: Author
    branch: str


@strawberry.type
class Magazine:
    id: int
    name: str
    issue: int
    branch: str


@strawberry.type
class Library:
    id: int
    branch: str
    magazine: List[Magazine]
    book: List[Book]


Item = Union[Book, Magazine]
Storage = List[str]


authors = [
    Author(
        first_name="New",
        last_name="Relic",
    ),
    Author(
        first_name="Bob",
        last_name="Smith",
    ),
    Author(
        first_name="Leslie",
        last_name="Jones",
    ),
]

books = [
    Book(
        id=1,
        name="Python Agent: The Book",
        isbn="a-fake-isbn",
        author=authors[0],
        branch="riverside",
    ),
    Book(
        id=2,
        name="Ollies for O11y: A Sk8er's Guide to Observability",
        isbn="a-second-fake-isbn",
        author=authors[1],
        branch="downtown",
    ),
    Book(
        id=3,
        name="[Redacted]",
        isbn="a-third-fake-isbn",
        author=authors[2],
        branch="riverside",
    ),
]

magazines = [
    Magazine(id=1, name="Reli Updates Weekly", issue=1, branch="riverside"),
    Magazine(id=2, name="Reli: The Forgotten Years", issue=2, branch="downtown"),
    Magazine(id=3, name="Node Weekly", issue=1, branch="riverside"),
]


libraries = ["riverside", "downtown"]
libraries = [
    Library(
        id=i + 1,
        branch=branch,
        magazine=[m for m in magazines if m.branch == branch],
        book=[b for b in books if b.branch == branch],
    )
    for i, branch in enumerate(libraries)
]

storage = []


def resolve_hello():
    return "Hello!"


async def resolve_hello_async():
    return "Hello!"


def resolve_echo(echo: str):
    return echo


def resolve_library(index: int):
    return libraries[index]


def resolve_storage_add(string: str):
    storage.add(string)
    return storage


def resolve_storage():
    return storage


def resolve_error():
    raise RuntimeError("Runtime Error!")


def resolve_search(contains: str):
    search_books = [b for b in books if contains in b.name]
    search_magazines = [m for m in magazines if contains in m.name]
    return search_books + search_magazines


@strawberry.type
class Query:
    library: Library = field(resolver=resolve_library)
    hello: str = field(resolver=resolve_hello)
    hello_async: str = field(resolver=resolve_hello_async)
    search: List[Item] = field(resolver=resolve_search)
    echo: str = field(resolver=resolve_echo)
    storage: Storage = field(resolver=resolve_storage)
    error: Optional[str] = field(resolver=resolve_error)
    error_non_null: str = field(resolver=resolve_error)

    def resolve_library(self, info, index):
        return libraries[index]

    def resolve_storage(self, info):
        return storage

    def resolve_search(self, info, contains):
        search_books = [b for b in books if contains in b.name]
        search_magazines = [m for m in magazines if contains in m.name]
        return search_books + search_magazines

    def resolve_hello(self, info):
        return "Hello!"

    def resolve_echo(self, info, echo):
        return echo

    def resolve_error(self, info) -> str:
        raise RuntimeError("Runtime Error!")


@strawberry.type
class Mutation:
    @strawberry.mutation
    def storage_add(self, string: str) -> str:
        storage.append(string)
        return str(string)


_target_application = Schema(query=Query, mutation=Mutation, config=StrawberryConfig(auto_camel_case=False))
_target_asgi_application = GraphQL(_target_application)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import pytest
from testing_support.fixtures import dt_enabled
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_span_events import validate_span_events


def get_starlette_version():
    import starlette

    version = getattr(starlette, "__version__", "0.0.0").split(".")
    return tuple(int(x) for x in version)


@pytest.fixture(scope="session")
def target_application():
    import _test_graphql

    return _test_graphql.target_application


@dt_enabled
@pytest.mark.parametrize("endpoint", ("/async", "/sync"))
@pytest.mark.skipif(get_starlette_version() >= (0, 17), reason="Starlette GraphQL support dropped in v0.17.0")
def test_graphql_metrics_and_attrs(target_application, endpoint):
    from graphql import __version__ as version

    from newrelic.hooks.framework_graphene import framework_details

    FRAMEWORK_METRICS = [
        ("Python/Framework/Graphene/%s" % framework_details()[1], 1),
        ("Python/Framework/GraphQL/%s" % version, 1),
    ]
    _test_scoped_metrics = [
        ("GraphQL/resolve/Graphene/hello", 1),
        ("GraphQL/operation/Graphene/query/<anonymous>/hello", 1),
    ]
    _test_unscoped_metrics = [
        ("GraphQL/all", 1),
        ("GraphQL/Graphene/all", 1),
        ("GraphQL/allWeb", 1),
        ("GraphQL/Graphene/allWeb", 1),
    ] + _test_scoped_metrics

    _expected_query_operation_attributes = {
        "graphql.operation.type": "query",
        "graphql.operation.name": "<anonymous>",
        "graphql.operation.query": "{ hello }",
    }
    _expected_query_resolver_attributes = {
        "graphql.field.name": "hello",
        "graphql.field.parentType": "Query",
        "graphql.field.path": "hello",
        "graphql.field.returnType": "String",
    }

    @validate_span_events(exact_agents=_expected_query_operation_attributes)
    @validate_span_events(exact_agents=_expected_query_resolver_attributes)
    @validate_transaction_metrics(
        "query/<anonymous>/hello",
        "GraphQL",
        scoped_metrics=_test_scoped_metrics,
        rollup_metrics=_test_unscoped_metrics + FRAMEWORK_METRICS,
    )
    def _test():
        response = target_application.make_request(
            "POST", endpoint, body=json.dumps({"query": "{ hello }"}), headers={"Content-Type": "application/json"}
        )
        assert response.status == 200
        assert "Hello!" in response.body.decode("utf-8")

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import uuid

import botocore.session
import moto
import pytest
from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_span_events import validate_span_events
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.common.package_version_utils import get_package_version

MOTO_VERSION = tuple(int(v) for v in moto.__version__.split(".")[:3])

# patch earlier versions of moto to support py37
if sys.version_info >= (3, 7) and MOTO_VERSION <= (1, 3, 1):
    import re

    moto.packages.responses.responses.re._pattern_type = re.Pattern

url = "sqs.us-east-1.amazonaws.com"
botocore_version = tuple([int(n) for n in get_package_version("botocore").split(".")])
if botocore_version < (1, 29, 0):
    url = "queue.amazonaws.com"

AWS_ACCESS_KEY_ID = "AAAAAAAAAAAACCESSKEY"
AWS_SECRET_ACCESS_KEY = "AAAAAASECRETKEY"  # nosec
AWS_REGION = "us-east-1"

TEST_QUEUE = "python-agent-test-%s" % uuid.uuid4()


_sqs_scoped_metrics = [
    ("MessageBroker/SQS/Queue/Produce/Named/%s" % TEST_QUEUE, 2),
    ("External/%s/botocore/POST" % url, 3),
]

_sqs_rollup_metrics = [
    ("MessageBroker/SQS/Queue/Produce/Named/%s" % TEST_QUEUE, 2),
    ("MessageBroker/SQS/Queue/Consume/Named/%s" % TEST_QUEUE, 1),
    ("External/all", 3),
    ("External/allOther", 3),
    ("External/%s/all" % url, 3),
    ("External/%s/botocore/POST" % url, 3),
]

_sqs_scoped_metrics_malformed = [
    ("MessageBroker/SQS/Queue/Produce/Named/Unknown", 1),
]

_sqs_rollup_metrics_malformed = [
    ("MessageBroker/SQS/Queue/Produce/Named/Unknown", 1),
]


@override_application_settings({"distributed_tracing.enabled": True})
@validate_span_events(exact_agents={"aws.operation": "CreateQueue"}, count=1)
@validate_span_events(exact_agents={"aws.operation": "SendMessage"}, count=1)
@validate_span_events(exact_agents={"aws.operation": "ReceiveMessage"}, count=1)
@validate_span_events(exact_agents={"aws.operation": "SendMessageBatch"}, count=1)
@validate_span_events(exact_agents={"aws.operation": "PurgeQueue"}, count=1)
@validate_span_events(exact_agents={"aws.operation": "DeleteQueue"}, count=1)
@validate_transaction_metrics(
    "test_botocore_sqs:test_sqs",
    scoped_metrics=_sqs_scoped_metrics,
    rollup_metrics=_sqs_rollup_metrics,
    background_task=True,
)
@background_task()
@moto.mock_sqs
def test_sqs():
    session = botocore.session.get_session()
    client = session.create_client(
        "sqs", region_name=AWS_REGION, aws_access_key_id=AWS_ACCESS_KEY_ID, aws_secret_access_key=AWS_SECRET_ACCESS_KEY
    )

    # Create queue
    resp = client.create_queue(QueueName=TEST_QUEUE)
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200

    # QueueUrl is needed for rest of methods.
    QUEUE_URL = resp["QueueUrl"]

    # Send message
    resp = client.send_message(QueueUrl=QUEUE_URL, MessageBody="hello_world")
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200

    # Receive message
    resp = client.receive_message(QueueUrl=QUEUE_URL)
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200

    # Send message batch
    messages = [
        {"Id": "1", "MessageBody": "message 1"},
        {"Id": "2", "MessageBody": "message 2"},
        {"Id": "3", "MessageBody": "message 3"},
    ]
    resp = client.send_message_batch(QueueUrl=QUEUE_URL, Entries=messages)
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200

    # Purge queue
    resp = client.purge_queue(QueueUrl=QUEUE_URL)
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200

    # Delete queue
    resp = client.delete_queue(QueueUrl=QUEUE_URL)
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200


@override_application_settings({"distributed_tracing.enabled": True})
@validate_transaction_metrics(
    "test_botocore_sqs:test_sqs_malformed",
    scoped_metrics=_sqs_scoped_metrics_malformed,
    rollup_metrics=_sqs_rollup_metrics_malformed,
    background_task=True,
)
@background_task()
@moto.mock_sqs
def test_sqs_malformed():
    session = botocore.session.get_session()
    client = session.create_client(
        "sqs", region_name=AWS_REGION, aws_access_key_id=AWS_ACCESS_KEY_ID, aws_secret_access_key=AWS_SECRET_ACCESS_KEY
    )

    # Malformed send message, uses arg instead of kwarg
    with pytest.raises(TypeError):
        client.send_message("https://fake-url/", MessageBody="hello_world")



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio


class Application:

    def __init__(self, scope):
        if scope['type'] != 'http':
            raise ValueError('Unsupported', scope['type'])
        self.scope = scope

    async def __call__(self, receive, send):
        await send({
            'type': 'http.response.start',
            'status': 200,
            'headers': [(b'Content-Type', b'text/plain')],
        })

        await send({
            'type': 'http.response.body',
            'body': b'PONG',
        })



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.asgi_application import ASGIApplicationWrapper
from newrelic.api.time_trace import notice_error
from newrelic.api.transaction import (
    add_custom_attribute,
    current_transaction,
    ignore_transaction,
)


class simple_app_v2_raw:
    def __init__(self, scope):
        self.scope = scope

    async def __call__(self, receive, send):
        if self.scope["type"] == "lifespan":
            return await handle_lifespan(self.scope, receive, send)

        if self.scope["type"] != "http":
            raise ValueError("unsupported")

        if self.scope["path"] == "/exc":
            raise ValueError("whoopsies")
        elif self.scope["path"] == "/ignored":
            ignore_transaction()

        await send({"type": "http.response.start", "status": 200})
        await send({"type": "http.response.body"})

        assert current_transaction() is None


class simple_app_v2_init_exc(simple_app_v2_raw):
    def __init__(self, scope):
        raise ValueError("oh no!")


async def simple_app_v3_raw(scope, receive, send):
    if scope["type"] == "lifespan":
        return await handle_lifespan(scope, receive, send)

    if scope["type"] != "http":
        raise ValueError("unsupported")

    if scope["path"] == "/exc":
        raise ValueError("whoopsies")
    elif scope["path"] == "/ignored":
        ignore_transaction()

    await send({"type": "http.response.start", "status": 200})
    await send({"type": "http.response.body"})

    assert current_transaction() is None


class AppWithDescriptor:
    @ASGIApplicationWrapper
    async def method(self, scope, receive, send):
        return await simple_app_v3_raw(scope, receive, send)

    @ASGIApplicationWrapper
    @classmethod
    async def cls(cls, scope, receive, send):
        return await simple_app_v3_raw(scope, receive, send)

    @ASGIApplicationWrapper
    @staticmethod
    async def static(scope, receive, send):
        return await simple_app_v3_raw(scope, receive, send)


class AppWithCallRaw:
    async def __call__(self, scope, receive, send):
        return await simple_app_v3_raw(scope, receive, send)


class AppWithCall(AppWithCallRaw):
    @ASGIApplicationWrapper
    async def __call__(self, scope, receive, send):
        return await super(AppWithCall, self).__call__(scope, receive, send)


simple_app_v2 = ASGIApplicationWrapper(simple_app_v2_raw)
simple_app_v2_init_exc = ASGIApplicationWrapper(simple_app_v2_init_exc)
simple_app_v3 = ASGIApplicationWrapper(simple_app_v3_raw)


@ASGIApplicationWrapper
async def normal_asgi_application(scope, receive, send):
    output = b"<html><head>header</head><body><p>RESPONSE</p></body></html>"
    add_custom_attribute("puppies", "test_value")
    add_custom_attribute("sunshine", "test_value")

    response_headers = [
        (b"content-type", b"text/html; charset=utf-8"),
        (b"content-length", str(len(output)).encode("utf-8")),
    ]

    try:
        raise ValueError("Transaction had bad value")
    except ValueError:
        notice_error(attributes={"ohnoes": "param-value"})

    await send({"type": "http.response.start", "status": 200, "headers": response_headers})
    await send({"type": "http.response.body", "body": output})


async def handle_lifespan(scope, receive, send):
    """Handle lifespan protocol with no-ops to allow more compatibility."""
    while True:
        txn = current_transaction()
        if txn:
            txn.ignore_transaction = True

        message = await receive()
        if message["type"] == "lifespan.startup":
            await send({"type": "lifespan.startup.complete"})
        elif message["type"] == "lifespan.shutdown":
            await send({"type": "lifespan.shutdown.complete"})
            return



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.transaction import current_transaction
from newrelic.common.encoding_utils import deobfuscate, json_decode


OUTBOUND_TRACE_KEYS_REQUIRED = (
        'ty', 'ac', 'ap', 'tr', 'pr', 'sa', 'ti')


def validate_outbound_headers(header_id='X-NewRelic-ID',
        header_transaction='X-NewRelic-Transaction'):
    transaction = current_transaction()
    headers = transaction._test_request_headers
    settings = transaction.settings
    encoding_key = settings.encoding_key

    assert header_id in headers

    values = headers[header_id]
    if isinstance(values, list):
        assert len(values) == 1, headers
        assert isinstance(values[0], type(''))
        value = values[0]
    else:
        value = values

    cross_process_id = deobfuscate(value, encoding_key)
    assert cross_process_id == settings.cross_process_id

    assert header_transaction in headers

    values = headers[header_transaction]
    if isinstance(values, list):
        assert len(values) == 1, headers
        assert isinstance(values[0], type(''))
        value = values[0]
    else:
        value = values

    (guid, record_tt, trip_id, path_hash) = \
            json_decode(deobfuscate(value, encoding_key))

    assert guid == transaction.guid
    assert record_tt == transaction.record_tt
    assert trip_id == transaction.trip_id
    assert path_hash == transaction.path_hash



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

import pytest
from testing_support.validators.validate_function_called import validate_function_called

from newrelic.common.package_version_utils import (
    NULL_VERSIONS,
    VERSION_ATTRS,
    get_package_version,
    get_package_version_tuple,
)

IS_PY38_PLUS = sys.version_info[:2] >= (3, 8)
SKIP_IF_NOT_IMPORTLIB_METADATA = pytest.mark.skipif(not IS_PY38_PLUS, reason="importlib.metadata is not supported.")
SKIP_IF_IMPORTLIB_METADATA = pytest.mark.skipif(
    IS_PY38_PLUS, reason="importlib.metadata is preferred over pkg_resources."
)


@pytest.fixture(scope="function", autouse=True)
def patched_pytest_module(monkeypatch):
    for attr in VERSION_ATTRS:
        if hasattr(pytest, attr):
            monkeypatch.delattr(pytest, attr)

    yield pytest


@pytest.mark.parametrize(
    "attr,value,expected_value",
    (
        ("version", "1.2.3.4", "1.2.3.4"),
        ("__version__", "1.3.5rc2", "1.3.5rc2"),
        ("__version_tuple__", (3, 5, 8), "3.5.8"),
        ("version_tuple", [3, 1, "0b2"], "3.1.0b2"),
    ),
)
def test_get_package_version(attr, value, expected_value):
    # There is no file/module here, so we monkeypatch
    # pytest instead for our purposes
    setattr(pytest, attr, value)
    version = get_package_version("pytest")
    assert version == expected_value
    delattr(pytest, attr)


@pytest.mark.parametrize(
    "attr,value,expected_value",
    (
        ("version", "1.2.3.4", (1, 2, 3, 4)),
        ("__version__", "1.3.5rc2", (1, 3, "5rc2")),
        ("__version_tuple__", (3, 5, 8), (3, 5, 8)),
        ("version_tuple", [3, 1, "0b2"], (3, 1, "0b2")),
    ),
)
def test_get_package_version_tuple(attr, value, expected_value):
    # There is no file/module here, so we monkeypatch
    # pytest instead for our purposes
    setattr(pytest, attr, value)
    version = get_package_version_tuple("pytest")
    assert version == expected_value
    delattr(pytest, attr)


@SKIP_IF_NOT_IMPORTLIB_METADATA
@validate_function_called("importlib.metadata", "version")
def test_importlib_metadata():
    version = get_package_version("pytest")
    assert version not in NULL_VERSIONS, version


@SKIP_IF_IMPORTLIB_METADATA
@validate_function_called("pkg_resources", "get_distribution")
def test_pkg_resources_metadata():
    version = get_package_version("pytest")
    assert version not in NULL_VERSIONS, version



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, ES_VERSION 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from conftest import ES_SETTINGS, ES_VERSION
from elasticsearch.serializer import JSONSerializer

from newrelic.api.background_task import background_task
from newrelic.api.transaction import current_transaction

try:
    from elasticsearch.connection.http_requests import RequestsHttpConnection
    from elasticsearch.connection.http_urllib3 import Urllib3HttpConnection
    from elasticsearch.transport import Transport

    NodeConfig = dict
except ImportError:
    from elastic_transport._models import NodeConfig
    from elastic_transport._node._http_requests import (
        RequestsHttpNode as RequestsHttpConnection,
    )
    from elastic_transport._node._http_urllib3 import (
        Urllib3HttpNode as Urllib3HttpConnection,
    )
    from elastic_transport._transport import Transport


IS_V8 = ES_VERSION >= (8,)
IS_V7 = ES_VERSION >= (7,) and ES_VERSION < (8, 0)
IS_BELOW_V7 = ES_VERSION < (7,)

RUN_IF_V8 = pytest.mark.skipif(IS_V7 or IS_BELOW_V7, reason="Only run for v8+")
RUN_IF_V7 = pytest.mark.skipif(IS_V8 or IS_BELOW_V7, reason="Only run for v7")
RUN_IF_BELOW_V7 = pytest.mark.skipif(not IS_BELOW_V7, reason="Only run for versions below v7")


HOST = NodeConfig(scheme="http", host=ES_SETTINGS["host"], port=int(ES_SETTINGS["port"]))

METHOD = "/contacts/person/1"
HEADERS = {"Content-Type": "application/json"}
DATA = {"name": "Joe Tester"}

BODY = JSONSerializer().dumps(DATA)
if hasattr(BODY, "encode"):
    BODY = BODY.encode("utf-8")


@pytest.mark.parametrize(
    "transport_kwargs, perform_request_kwargs",
    [
        pytest.param({}, {"body": DATA}, id="DefaultTransport_below_v7", marks=RUN_IF_BELOW_V7),
        pytest.param({}, {"headers": HEADERS, "body": DATA}, id="DefaultTransport_v7+", marks=RUN_IF_V7 or RUN_IF_V8),
        pytest.param(
            {"connection_class": Urllib3HttpConnection},
            {"body": DATA},
            id="Urllib3HttpConnectionv7",
            marks=RUN_IF_BELOW_V7,
        ),
        pytest.param(
            {"connection_class": RequestsHttpConnection},
            {"body": DATA},
            id="RequestsHttpConnectionv7",
            marks=RUN_IF_BELOW_V7,
        ),
        pytest.param(
            {"connection_class": Urllib3HttpConnection},
            {"headers": HEADERS, "body": DATA},
            id="Urllib3HttpConnectionv7",
            marks=RUN_IF_V7,
        ),
        pytest.param(
            {"connection_class": RequestsHttpConnection},
            {"headers": HEADERS, "body": DATA},
            id="RequestsHttpConnectionv7",
            marks=RUN_IF_V7,
        ),
        pytest.param(
            {"node_class": Urllib3HttpConnection},
            {"headers": HEADERS, "body": DATA},
            id="Urllib3HttpNodev8",
            marks=RUN_IF_V8,
        ),
        pytest.param(
            {"node_class": RequestsHttpConnection},
            {"headers": HEADERS, "body": DATA},
            id="RequestsHttpNodev8",
            marks=RUN_IF_V8,
        ),
    ],
)
@background_task()
def test_transport_connection_classes(transport_kwargs, perform_request_kwargs):
    transaction = current_transaction()

    transport = Transport([HOST], **transport_kwargs)
    transport.perform_request("POST", METHOD, **perform_request_kwargs)

    expected = (ES_SETTINGS["host"], ES_SETTINGS["port"], None)
    assert transaction._nr_datastore_instance_info == expected



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import uuid

import botocore.session
import moto

from newrelic.api.background_task import background_task
from testing_support.fixtures import (
        validate_tt_segment_params, override_application_settings)
from testing_support.validators.validate_span_events import (
        validate_span_events)
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics

MOTO_VERSION = tuple(int(v) for v in moto.__version__.split('.')[:3])

# patch earlier versions of moto to support py37
if sys.version_info >= (3, 7) and MOTO_VERSION <= (1, 3, 1):
    import re
    moto.packages.responses.responses.re._pattern_type = re.Pattern

AWS_ACCESS_KEY_ID = 'AAAAAAAAAAAACCESSKEY'
AWS_SECRET_ACCESS_KEY = 'AAAAAASECRETKEY'
AWS_REGION = 'us-east-1'
UBUNTU_14_04_PARAVIRTUAL_AMI = 'ami-c65be9ae'

TEST_INSTANCE = 'python-agent-test-%s' % uuid.uuid4()

_ec2_scoped_metrics = [
    ('External/ec2.us-east-1.amazonaws.com/botocore/POST', 3),
]

_ec2_rollup_metrics = [
    ('External/all', 3),
    ('External/allOther', 3),
    ('External/ec2.us-east-1.amazonaws.com/all', 3),
    ('External/ec2.us-east-1.amazonaws.com/botocore/POST', 3),
]


@override_application_settings({'distributed_tracing.enabled': True})
@validate_span_events(expected_agents=('aws.requestId',), count=3)
@validate_span_events(exact_agents={'aws.operation': 'RunInstances'}, count=1)
@validate_span_events(
        exact_agents={'aws.operation': 'DescribeInstances'}, count=1)
@validate_span_events(
        exact_agents={'aws.operation': 'TerminateInstances'}, count=1)
@validate_tt_segment_params(present_params=('aws.requestId',))
@validate_transaction_metrics(
        'test_botocore_ec2:test_ec2',
        scoped_metrics=_ec2_scoped_metrics,
        rollup_metrics=_ec2_rollup_metrics,
        background_task=True)
@background_task()
@moto.mock_ec2
def test_ec2():
    session = botocore.session.get_session()
    client = session.create_client(
            'ec2',
            region_name=AWS_REGION,
            aws_access_key_id=AWS_ACCESS_KEY_ID,
            aws_secret_access_key=AWS_SECRET_ACCESS_KEY
    )

    # Create instance
    resp = client.run_instances(
            ImageId=UBUNTU_14_04_PARAVIRTUAL_AMI,
            InstanceType='m1.small',
            MinCount=1,
            MaxCount=1,
    )
    assert resp['ResponseMetadata']['HTTPStatusCode'] == 200
    assert len(resp['Instances']) == 1
    instance_id = resp['Instances'][0]['InstanceId']

    # Describe instance
    resp = client.describe_instances(InstanceIds=[instance_id])
    assert resp['ResponseMetadata']['HTTPStatusCode'] == 200
    assert resp['Reservations'][0]['Instances'][0]['InstanceId'] == instance_id

    # Delete instance
    resp = client.terminate_instances(InstanceIds=[instance_id])
    assert resp['ResponseMetadata']['HTTPStatusCode'] == 200
    assert resp['TerminatingInstances'][0]['InstanceId'] == instance_id



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from solr import SolrConnection

from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.db_settings import solr_settings

from newrelic.api.background_task import background_task

DB_SETTINGS = solr_settings()[0]
SOLR_HOST = DB_SETTINGS["host"]
SOLR_PORT = DB_SETTINGS["port"]
SOLR_URL = 'http://%s:%s/solr/collection' % (DB_SETTINGS["host"], DB_SETTINGS["port"])

def _exercise_solr(solr):
    # Construct document names within namespace
    documents = ["pysolr_doc_1", "pysolr_doc_2"]
    documents = [x + "_" + DB_SETTINGS["namespace"] for x in documents]

    solr.add_many([{"id": x} for x in documents])
    solr.commit()
    solr.query('id:%s' % documents[0]).results
    solr.delete('id:*_%s' % DB_SETTINGS["namespace"])
    solr.commit()

_test_solr_search_scoped_metrics = [
        ('Datastore/operation/Solr/add_many', 1),
        ('Datastore/operation/Solr/delete', 1),
        ('Datastore/operation/Solr/commit', 2),
        ('Datastore/operation/Solr/query', 1)]

_test_solr_search_rollup_metrics = [
        ('Datastore/all', 5),
        ('Datastore/allOther', 5),
        ('Datastore/Solr/all', 5),
        ('Datastore/Solr/allOther', 5),
        ('Datastore/operation/Solr/add_many', 1),
        ('Datastore/operation/Solr/query', 1),
        ('Datastore/operation/Solr/commit', 2),
        ('Datastore/operation/Solr/delete', 1)]

@validate_transaction_metrics('test_solr:test_solr_search',
    scoped_metrics=_test_solr_search_scoped_metrics,
    rollup_metrics=_test_solr_search_rollup_metrics,
    background_task=True)
@background_task()
def test_solr_search():
    s = SolrConnection(SOLR_URL)
    _exercise_solr(s)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

# import aioredis
from conftest import AIOREDIS_VERSION, loop  # noqa # pylint: disable=E0611,W0611
from testing_support.db_settings import redis_settings
from testing_support.fixtures import override_application_settings
from testing_support.util import instance_hostname
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task

DB_SETTINGS = redis_settings()[0]

SKIP_IF_AIOREDIS_V1 = pytest.mark.skipif(AIOREDIS_VERSION < (2, 0), reason="Single arg commands not supported.")

_enable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": True,
}

_disable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": False,
}

_base_scoped_metrics = (("Datastore/operation/Redis/client_list", 1),)

_base_rollup_metrics = (
    ("Datastore/all", 1),
    ("Datastore/allOther", 1),
    ("Datastore/Redis/all", 1),
    ("Datastore/Redis/allOther", 1),
    ("Datastore/operation/Redis/client_list", 1),
)

_disable_scoped_metrics = list(_base_scoped_metrics)
_disable_rollup_metrics = list(_base_rollup_metrics)

_enable_scoped_metrics = list(_base_scoped_metrics)
_enable_rollup_metrics = list(_base_rollup_metrics)

_host = instance_hostname(DB_SETTINGS["host"])
_port = DB_SETTINGS["port"]

_instance_metric_name = "Datastore/instance/Redis/%s/%s" % (_host, _port)

_enable_rollup_metrics.append((_instance_metric_name, 1))

_disable_rollup_metrics.append((_instance_metric_name, None))


async def exercise_redis_multi_args(client):
    if hasattr(client, "execute_command"):
        await client.execute_command("CLIENT", "LIST", parse="LIST")
    else:
        await client.execute("CLIENT", "LIST")


async def exercise_redis_single_arg(client):
    await client.execute_command("CLIENT LIST")


@SKIP_IF_AIOREDIS_V1
@override_application_settings(_enable_instance_settings)
@validate_transaction_metrics(
    "test_execute_command:test_redis_execute_command_as_one_arg_enable",
    scoped_metrics=_enable_scoped_metrics,
    rollup_metrics=_enable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_redis_execute_command_as_one_arg_enable(client, loop):  # noqa
    loop.run_until_complete(exercise_redis_single_arg(client))


@SKIP_IF_AIOREDIS_V1
@override_application_settings(_disable_instance_settings)
@validate_transaction_metrics(
    "test_execute_command:test_redis_execute_command_as_one_arg_disable",
    scoped_metrics=_disable_scoped_metrics,
    rollup_metrics=_disable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_redis_execute_command_as_one_arg_disable(client, loop):  # noqa
    loop.run_until_complete(exercise_redis_single_arg(client))


@override_application_settings(_enable_instance_settings)
@validate_transaction_metrics(
    "test_execute_command:test_redis_execute_command_as_two_args_enable",
    scoped_metrics=_enable_scoped_metrics,
    rollup_metrics=_enable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_redis_execute_command_as_two_args_enable(client, loop):  # noqa
    loop.run_until_complete(exercise_redis_multi_args(client))


@override_application_settings(_disable_instance_settings)
@validate_transaction_metrics(
    "test_execute_command:test_redis_execute_command_as_two_args_disable",
    scoped_metrics=_disable_scoped_metrics,
    rollup_metrics=_disable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_redis_execute_command_as_two_args_disable(client, loop):  # noqa
    loop.run_until_complete(exercise_redis_multi_args(client))
